import abc
import dask.dataframe as dd
import functools as ft
import inflect
import numpy as np
import pandas as pd
from dask.delayed import Delayed
from dask.utils import format_bytes
from enum import Enum
from jinja2 import Environment
from pathlib import Path
from scipy.cluster.hierarchy import linkage, fcluster
from typing import Dict, List, Union

from .analysis_utils import set_file_dir, set_file_pattern, set_proc_name_parts
from .constants import (
    ACC_PAT_SUFFIXES,
    COL_APP_NAME,
    COL_COUNT,
    COL_FILE_DIR,
    COL_FILE_NAME,
    COL_FILE_PATTERN,
    COL_NODE_NAME,
    COL_PROC_NAME,
    COL_RANK,
    COL_TIME_RANGE,
    COMPACT_IO_TYPES,
    HUMANIZED_VIEW_TYPES,
    IO_TYPES,
    XFER_SIZE_BINS,
    XFER_SIZE_BIN_LABELS,
    XFER_SIZE_BIN_NAMES,
    AccessPattern,
)
from .scoring import SCORING_ORDER
from .types import (
    BottleneckResult,
    Metric,
    RawStats,
    Rule,
    RuleReason,
    RuleResult,
    RuleResultReason,
    ScoringResult,
    ViewKey,
    ViewResult,
)
from .utils.collection_utils import get_intervals, join_with_and
from .utils.common_utils import format_number, numerize


MAX_REASONS = 5
METADATA_ACCESS_RATIO_THRESHOLD = 0.5


jinja_env = Environment()
jinja_env.filters['format_bytes'] = format_bytes
jinja_env.filters['format_number'] = format_number


class KnownCharacteristics(Enum):
    ACCESS_PATTERN = 'access_pattern'
    APP_COUNT = 'app_count'
    COMPLEXITY = 'complexity'
    FILE_COUNT = 'file_count'
    IO_COUNT = 'io_count'
    IO_SIZE = 'io_size'
    IO_TIME = 'io_time'
    NODE_COUNT = 'node_count'
    PROC_COUNT = 'proc_count'
    READ_XFER_SIZE = 'read_xfer_size'
    TIME_PERIOD = 'time_period'
    WRITE_XFER_SIZE = 'write_xfer_size'


class KnownRules(Enum):
    EXCESSIVE_METADATA_ACCESS = 'excessive_metadata_access'
    OPERATION_IMBALANCE = 'operation_imbalance'
    RANDOM_OPERATIONS = 'random_operations'
    SIZE_IMBALANCE = 'size_imbalance'
    SMALL_READS = 'small_reads'
    SMALL_WRITES = 'small_writes'


KNOWN_RULES = {
    KnownRules.EXCESSIVE_METADATA_ACCESS.value: Rule(
        name='Excessive metadata access',
        condition='(metadata_time / time) >= 0.5',
        reasons=[
            RuleReason(
                condition='(open_time > close_time) & (open_time > seek_time)',
                message='''
Overall {{ "%.2f" | format((metadata_time / time) * 100) }}% ({{ "%.2f" | format(metadata_time) }} seconds) of I/O time is spent on metadata access, \
specifically {{ "%.2f" | format((open_time / time) * 100) }}% ({{ "%.2f" | format(open_time) }} seconds) on the 'open' operation.
                ''',
            ),
            RuleReason(
                condition='(close_time > open_time) & (close_time > seek_time)',
                message='''
Overall {{ "%.2f" | format((metadata_time / time) * 100) }}% ({{ "%.2f" | format(metadata_time) }} seconds) of I/O time is spent on metadata access, \
specifically {{ "%.2f" | format((open_time / time) * 100) }}% ({{ "%.2f" | format(open_time) }} seconds) on the 'close' operation.
                ''',
            ),
            RuleReason(
                condition='(seek_time > open_time) & (seek_time > close_time)',
                message='''
Overall {{ "%.2f" | format((metadata_time / time) * 100) }}% ({{ "%.2f" | format(metadata_time) }} seconds) of I/O time is spent on metadata access, \
specifically {{ "%.2f" | format((open_time / time) * 100) }}% ({{ "%.2f" | format(open_time) }} seconds) on the 'seek' operation.
                ''',
            ),
        ],
    ),
    KnownRules.OPERATION_IMBALANCE.value: Rule(
        name='Operation imbalance',
        condition='(abs(write_count - read_count) / count) > 0.1',
        reasons=[
            RuleReason(
                condition='read_count > write_count',
                message='''
'read' operations are {{ "%.2f" | format((read_count / count) * 100) }}% ({{ read_count | format_number }} operations) of total I/O operations.
                ''',
            ),
            RuleReason(
                condition='write_count > read_count',
                message='''
'write' operations are {{ "%.2f" | format((write_count / count) * 100) }}% ({{ write_count | format_number }} operations) of total I/O operations.
                ''',
            ),
        ],
    ),
    KnownRules.RANDOM_OPERATIONS.value: Rule(
        name='Random operations',
        condition='random_count / count > 0.5',
        reasons=[
            RuleReason(
                condition='random_count / count > 0.5',
                message='''
Issued high number of random operations, specifically {{ "%.2f" | format((random_count / count) * 100) }}% \
({{ random_count | format_number }} operations) of total I/O operations.
                ''',
            ),
        ],
    ),
    KnownRules.SIZE_IMBALANCE.value: Rule(
        name='Size imbalance',
        condition='size > 0 & (abs(write_size - read_size) / size) > 0.1',
        reasons=[
            RuleReason(
                condition='read_size > write_size',
                message='''
'read' size is {{ "%.2f" | format((read_size / size) * 100) }}% ({{ read_size | format_bytes }}) of total I/O size.
                ''',
            ),
            RuleReason(
                condition='write_size > read_size',
                message='''
'write' size is {{ "%.2f" | format((write_size / size) * 100) }}% ({{ write_size | format_bytes }}) of total I/O size.
                ''',
            ),
        ],
    ),
    KnownRules.SMALL_READS.value: Rule(
        name='Small reads',
        condition='(read_time / time) > 0.5 & (read_size / count) < 1048576',
        reasons=[
            RuleReason(
                condition='(read_time / time) > 0.5',
                message='''
'read' time is {{ "%.2f" | format((read_time / time) * 100) }}% ({{ "%.2f" | format(read_time) }} seconds) of I/O time.
                ''',
            ),
            RuleReason(
                condition='(read_size / count) < 1048576',
                message='''
Average 'read's are {{ (read_size / count) | format_bytes }}, which is smaller than {{ 1048576 | format_bytes }}.
                ''',
            ),
        ],
    ),
    KnownRules.SMALL_WRITES.value: Rule(
        name='Small writes',
        condition='(write_time / time) > 0.5 & (write_size / count) < 1048576',
        reasons=[
            RuleReason(
                condition='(write_time / time) > 0.5',
                message='''
'write' time is {{ "%.2f" | format((write_time / time) * 100) }}% ({{ "%.2f" | format(write_time) }} seconds) of I/O time.
                ''',
            ),
            RuleReason(
                condition='(write_size / count) < 1048576',
                message='''
Average 'write's are {{ (write_size / count) | format_bytes }}, which is smaller than {{ 1048576 | format_bytes }}.
                ''',
            ),
        ],
    ),
}

HUMANIZED_KNOWN_RULES = {rule: KNOWN_RULES[rule].name for rule in KNOWN_RULES}


class RuleHandler(abc.ABC):
    def __init__(self, rule_key: str) -> None:
        super().__init__()
        self.pluralize = inflect.engine()
        self.rule_key = rule_key


class BottleneckRule(RuleHandler):
    def __init__(self, rule_key: str, rule: Rule, verbose: bool = False) -> None:
        super().__init__(rule_key=rule_key)
        self.rule = rule
        self.verbose = verbose

    def define_tasks(
        self,
        metric: Metric,
        metric_boundary: dd.core.Scalar,
        scoring_result: ScoringResult,
        view_key: ViewKey,
    ) -> Dict[str, Delayed]:
        view_type = view_key[-1]

        bottlenecks = scoring_result.scored_view.query(self.rule.condition)
        bottlenecks['time_overall'] = bottlenecks['time'] / metric_boundary

        details = (
            scoring_result.records_index.to_frame()
            .reset_index(drop=True)
            .query(f"{view_type} in @indices", local_dict={'indices': bottlenecks.index})
        )

        tasks = {}
        tasks['bottlenecks'] = bottlenecks
        tasks['details'] = details

        for i, reason in enumerate(self.rule.reasons):
            tasks[f"reason{i}"] = bottlenecks.eval(reason.condition)

        return tasks

    def handle_task_results(
        self,
        metric: Metric,
        view_key: ViewKey,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series, pd.Index]],
    ) -> List[RuleResult]:
        # t0 = time.perf_counter()

        view_type = view_key[-1]

        bottlenecks = result['bottlenecks']

        if len(bottlenecks) == 0:
            return []

        details = result['details'].to_frame(index=False)
        metric_boundary = result['metric_boundary']

        files = {}
        processes = {}
        time_periods = {}

        num_files = {}
        num_ops = bottlenecks[COL_COUNT].to_dict()
        num_processes = {}
        num_time_periods = {}

        # print('handle_task_results t0', time.perf_counter() - t0)

        # Logical view type fix
        if view_type == COL_FILE_DIR:
            details = set_file_dir(df=details.set_index(COL_FILE_NAME))
        elif view_type == COL_FILE_PATTERN:
            details = set_file_pattern(df=details.set_index(COL_FILE_NAME))
        elif view_type in [COL_APP_NAME, COL_NODE_NAME, COL_RANK]:
            details = set_proc_name_parts(df=details.set_index(COL_PROC_NAME))

        if self.verbose:
            for col in details.columns:
                if col in [COL_FILE_NAME, COL_FILE_DIR, COL_FILE_PATTERN]:
                    files = details.groupby(view_type)[col].unique().to_dict()
                    num_files = {f: len(files[f]) for f in files}
                if col in [COL_APP_NAME, COL_NODE_NAME, COL_PROC_NAME, COL_RANK]:
                    processes = details.groupby(view_type)[col].unique().to_dict()
                    num_processes = {p: len(processes[p]) for p in processes}
                if col == COL_TIME_RANGE:
                    time_periods = details.groupby(view_type)[col].unique().to_dict()
                    num_time_periods = {t: len(time_periods[t]) for t in time_periods}
        else:
            for col in details.columns:
                if col in [COL_FILE_NAME, COL_FILE_DIR, COL_FILE_PATTERN]:
                    num_files = details.groupby(view_type)[col].nunique().to_dict()
                if col in [COL_APP_NAME, COL_NODE_NAME, COL_PROC_NAME, COL_RANK]:
                    num_processes = details.groupby(view_type)[col].nunique().to_dict()
                if col == COL_TIME_RANGE:
                    num_time_periods = details.groupby(view_type)[col].nunique().to_dict()

        # print('handle_task_results t1', time.perf_counter() - t0)

        reasoning = {}
        reasoning_templates = {}
        for i, reason in enumerate(self.rule.reasons):
            reasoning[i] = result[f"reason{i}"].to_dict()
            reasoning_templates[i] = jinja_env.from_string(reason.message)

        results = []

        for row in bottlenecks.itertuples():
            bot_files = list(files.get(row.Index, []))
            bot_processes = list(processes.get(row.Index, []))
            bot_time_periods = list(time_periods.get(row.Index, []))

            bot_num_files = num_files.get(row.Index, 0)
            bot_num_ops = num_ops.get(row.Index, 0)
            bot_num_processes = num_processes.get(row.Index, 0)
            bot_num_time_periods = num_time_periods.get(row.Index, 0)

            description = self.describe_bottleneck(
                files=bot_files,
                subject=row.Index,
                metric=metric,
                metric_boundary=metric_boundary,
                num_files=bot_num_files,
                num_ops=bot_num_ops,
                num_processes=bot_num_processes,
                num_time_periods=bot_num_time_periods,
                processes=bot_processes,
                row=row,
                time_periods=bot_time_periods,
                view_type=view_type,
            )

            row_dict = row._asdict()

            if self.verbose:
                row_dict['files'] = bot_files
                row_dict['processes'] = bot_processes
                row_dict['time_periods'] = list(map(int, bot_time_periods))
                # print(row_dict)

            row_dict['num_files'] = bot_num_files
            row_dict['num_ops'] = bot_num_ops
            row_dict['num_processes'] = bot_num_processes
            row_dict['num_time_periods'] = bot_num_time_periods

            # Fix index
            row_dict[view_type] = row_dict['Index']
            del row_dict['Index']

            reasons = []
            for i, reason in enumerate(self.rule.reasons):
                if reasoning[i][row.Index]:
                    reasons.append(RuleResultReason(description=reasoning_templates[i].render(row_dict).strip()))

            if len(reasons) == 0:
                reasons.append(RuleResultReason(description='No reason found, further investigation needed.'))

            result = RuleResult(
                compact_desc=None,
                description=description,
                detail_list=None,
                extra_data=row_dict,
                object_hash=hash(
                    '_'.join(
                        [
                            '{:,.6f}'.format(row_dict['time']),
                            f"{row_dict['num_files']}",
                            f"{row_dict['num_processes']}",
                            f"{row_dict['num_time_periods']}",
                        ]
                    )
                ),
                reasons=reasons,
                value=row_dict['time'],
                value_fmt='{:,.6f}'.format(row_dict['time']),
            )

            results.append(result)

        # print('handle_task_results t3', time.perf_counter() - t0, len(results))

        return results

    def describe_bottleneck(
        self,
        metric: Metric,
        num_files: int,
        num_ops: int,
        num_processes: int,
        num_time_periods: int,
        subject: Union[str, int],
        time: float,
        time_overall: float,
        view_type: str,
        files=[],
        processes=[],
        time_periods=[],
        compact=False,
    ) -> str:
        if num_files > 0 and num_processes > 0 and num_time_periods > 0:
            nice_view_type = HUMANIZED_VIEW_TYPES[COL_PROC_NAME].lower()
            accessor_name = ' '
            accessed = HUMANIZED_VIEW_TYPES[COL_FILE_NAME].lower()
            accessed_name = ' '
            if view_type in [COL_FILE_NAME, COL_FILE_DIR, COL_FILE_PATTERN]:
                accessed = HUMANIZED_VIEW_TYPES[view_type].lower()
                if num_files == 1:
                    if compact:
                        accessed_name = f" ({Path(subject).name}) "
                    else:
                        accessed_name = f" ({subject}) "
            if view_type in [COL_APP_NAME, COL_NODE_NAME, COL_PROC_NAME, COL_RANK]:
                nice_view_type = HUMANIZED_VIEW_TYPES[view_type].lower()
                if num_processes == 1:
                    accessor_name = f" ({subject}) "

            accessor_noun = self.pluralize.plural_noun(nice_view_type, num_processes)
            accessor_verb = self.pluralize.plural_verb('accesses', num_processes)
            accessed_noun = self.pluralize.plural_noun(accessed, num_files)
            time_period_name = f" ({subject}) " if view_type == COL_TIME_RANGE else ' '
            time_period_noun = self.pluralize.plural_noun('time period', num_time_periods)

            if self.verbose:
                time_intervals = get_intervals(values=time_periods)

                description = (
                    f"{self.pluralize.join(processes)} {accessor_noun} {accessor_verb} "
                    f"{accessed_noun} {self.pluralize.join(files)} "
                    f"during the {join_with_and(values=time_intervals)}th {time_period_noun} "
                    f"and {self.pluralize.plural_verb('has', num_processes)} an I/O time of {time:.2f} seconds which is "
                    f"{time_overall * 100:.2f}% of overall I/O time of the workload."
                )
            else:
                # 32 processes access 1 file pattern within 6 time periods and have an I/O time of 2.92 seconds which
                # is 70.89% of overall I/O time of the workload.
                description = (
                    f"{num_processes:,} {accessor_noun}{accessor_name}{accessor_verb} "
                    f"{num_files:,} {accessed_noun}{accessed_name}"
                    f"within {num_time_periods:,} {time_period_noun}{time_period_name}"
                    f"across {num_ops:,} I/O {self.pluralize.plural_noun('operation', num_ops)} "
                    f"and {self.pluralize.plural_verb('has', num_processes)} an I/O time of {time:.2f} seconds which is "
                    f"{time_overall * 100:.2f}% of overall I/O time of the workload."
                )

        else:
            nice_subject = subject
            nice_view_type = HUMANIZED_VIEW_TYPES[view_type].lower()

            count = 1
            if view_type in [COL_FILE_NAME, COL_FILE_DIR, COL_FILE_PATTERN]:
                count = num_files
                if compact:
                    nice_subject = Path(subject).name
            elif view_type in [COL_APP_NAME, COL_NODE_NAME, COL_PROC_NAME, COL_RANK]:
                count = num_processes
            else:
                count = num_time_periods

            description = (
                f"{count:,} {self.pluralize.plural_noun(nice_view_type, count)} ({nice_subject}) "
                f"{self.pluralize.plural_verb('has', count)} an I/O time of {time:.2f} seconds "
                f"across {num_ops:,} I/O {self.pluralize.plural_noun('operation', num_ops)} "
                f"which is {time_overall * 100:.2f}% of overall I/O time of the workload."
            )

        return description

    def describe_reason(self, bottleneck: dict, reason_index: int):
        reason_template = self.rule.reasons[reason_index].message
        return jinja_env.from_string(reason_template).render(bottleneck).strip()

    @staticmethod
    def _group_similar_behavior(bottlenecks: pd.DataFrame, metric: str, view_type: str):
        behavior_col = 'behavior'
        cols = bottlenecks.columns

        if len(bottlenecks) > 1:
            behavior_cols = cols[cols.str.contains('_min|_max|_count|_size')]

            link_mat = linkage(bottlenecks[behavior_cols], method='single')
            behavior_labels = fcluster(link_mat, t=10, criterion='distance')

            bottlenecks[behavior_col] = behavior_labels
        else:
            bottlenecks[behavior_col] = 1

        agg_dict = {col: 'mean' for col in cols}
        agg_dict[view_type] = list
        agg_dict.pop(f"{metric}_score")

        return bottlenecks.reset_index().groupby(behavior_col).agg(agg_dict)

    @staticmethod
    def _union_details(details: pd.DataFrame, behavior: int, indices: list, view_type: str):
        view_types = details.index.names

        filtered_details = details.query(f"{view_type} in @indices", local_dict={'indices': indices})

        if len(view_types) == 1:
            # This means there is only one view type
            detail_groups = filtered_details

            # So override aggregations accordingly
            agg_dict = {}
            agg_dict[view_type] = set
        else:
            agg_dict = {col: set for col in SCORING_ORDER[view_type]}
            agg_dict.pop(view_type)

            for agg_key in agg_dict.copy():
                if agg_key not in view_types:
                    agg_dict.pop(agg_key)

            detail_groups = filtered_details.reset_index().groupby(view_type).agg(agg_dict)

            # Then create unions for other view types
            agg_dict = {col: lambda x: set.union(*x) for col in agg_dict}
            agg_dict[view_type] = set

        return (
            detail_groups.reset_index()
            .assign(behavior=behavior)
            .groupby(['behavior'])
            .agg(agg_dict)
            .reset_index(drop=True)
        )


class CharacteristicRule(RuleHandler):
    deps: List[str] = []

    @abc.abstractmethod
    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        raise NotImplementedError

    @abc.abstractmethod
    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        raise NotImplementedError


class CharacteristicAccessPatternRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.ACCESS_PATTERN.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        acc_pat_cols = []
        for acc_pat in list(AccessPattern):
            for col_suffix in ACC_PAT_SUFFIXES:
                col_name = f"{acc_pat.name.lower()}_{col_suffix}"
                acc_pat_cols.append(col_name)

        return {'acc_pat_sum': main_view[acc_pat_cols].sum()}

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        acc_pat_sum = result['acc_pat_sum']

        sequential_count = int(acc_pat_sum['sequential_count'])
        random_count = int(acc_pat_sum['random_count']) if 'random_count' in acc_pat_sum else 0
        total_count = sequential_count + random_count

        sequential_title = '[bold]Sequential[/bold]'
        random_title = '[bold]Random[/bold]'

        if total_count > 0:
            sequential_per_fmt = f"{sequential_count / total_count * 100:.2f}"
            random_per_fmt = f"{random_count / total_count * 100:.2f}"

            compact_desc = f"{sequential_title}: {sequential_per_fmt}% - {random_title}: {random_per_fmt}% "

            value_fmt = (
                f"{sequential_title}: {sequential_count:,} ops ({sequential_per_fmt}%) - "
                f"{random_title}: {random_count:,} ops ({random_per_fmt}%) "
            )
        else:
            compact_desc = f"{sequential_title}: N/A - {random_title}: N/A "

            value_fmt = f"{sequential_title}: N/A - {random_title}: N/A "

        return RuleResult(
            compact_desc=compact_desc,
            description='Access Pattern',
            detail_list=None,
            extra_data=dict(acc_pat_sum),
            object_hash=None,
            reasons=None,
            value=None,
            value_fmt=value_fmt,
        )


class CharacteristicComplexityRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.COMPLEXITY.value)
        self.deps = [
            KnownCharacteristics.FILE_COUNT.value,
            KnownCharacteristics.PROC_COUNT.value,
            KnownCharacteristics.TIME_PERIOD.value,
        ]

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        tasks = {}

        return tasks

    def handle_task_results(
        self,
        result: dict,
        characteristics: Dict[str, RuleResult],
        raw_stats: RawStats = None,
    ) -> RuleResult:
        num_files = characteristics[KnownCharacteristics.FILE_COUNT.value].value
        num_processes = characteristics[KnownCharacteristics.PROC_COUNT.value].value
        num_time_periods = characteristics[KnownCharacteristics.TIME_PERIOD.value].value

        complexities = np.array([num_processes, num_time_periods, num_files])
        complexity = np.log10(ft.reduce(np.multiply, complexities[complexities != 0]))

        return RuleResult(
            compact_desc=f"{complexity:.2f}",
            description='Complexity',
            detail_list=None,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=complexity,
            value_fmt=f"{complexity:.2f}",
        )


class CharacteristicFileCountRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.FILE_COUNT.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        x = main_view.reset_index()

        tasks = {}

        if COL_FILE_NAME in x.columns:
            tasks['total_count'] = x[COL_FILE_NAME].nunique()

            if COL_PROC_NAME in x.columns:
                fpp = x.groupby([COL_FILE_NAME])[COL_PROC_NAME].nunique().to_frame()

                fpp_count = fpp[fpp[COL_PROC_NAME] == 1][COL_PROC_NAME].count()

            else:
                fpp_count = 0

            tasks['fpp_count'] = fpp_count
        else:
            tasks['total_count'] = 0
            tasks['fpp_count'] = 0

        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        total_count = int(result['total_count'])
        fpp_count = int(result['fpp_count'])

        value_fmt = f"{total_count:,} {self.pluralize.plural_noun('file', total_count)}"

        compact_desc = [value_fmt]
        detail_list = []

        shared_title = 'Shared'
        fpp_title = 'FPP'

        if total_count == 0 or fpp_count == 0:
            detail_list.append(f"{shared_title}: N/A")
            detail_list.append(f"{fpp_title}: N/A")
        else:
            fpp_per = f"{fpp_count / total_count * 100:.2f}%"

            shared_count = total_count - fpp_count
            shared_per = f"{shared_count / total_count * 100:.2f}%"

            shared_fmt = f"{shared_count:,} {self.pluralize.plural_noun('file', shared_count)}"
            fpp_fmt = f"{fpp_count:,} {self.pluralize.plural_noun('file', fpp_count)}"

            compact_desc.append(f"[bold]{shared_title}[/bold]: {shared_per}")
            compact_desc.append(f"[bold]{fpp_title}[/bold]: {fpp_per}")

            detail_list.append(f"{shared_title}: {shared_fmt} ({shared_per})")
            detail_list.append(f"{fpp_title}: {fpp_fmt} ({fpp_per})")

        return RuleResult(
            compact_desc=' - '.join(compact_desc),
            description='Files',
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=total_count,
            value_fmt=value_fmt,
        )


class CharacteristicIOOpsRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.IO_COUNT.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        tasks = {}
        tasks['total_count'] = main_view['count'].sum()
        for io_type in IO_TYPES:
            count_col = f"{io_type}_count"
            tasks[count_col] = main_view[count_col].sum()
        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        total_count = int(result['total_count'])

        compact_desc = [f"{numerize(total_count, 1)} ops"]
        detail_list = []

        for i, io_type in enumerate(IO_TYPES):
            count_col = f"{io_type}_count"
            count = int(result[count_col])
            percent = f"{count / total_count * 100:.2f}%"
            compact_desc.append(f"[bold]{COMPACT_IO_TYPES[i]}[/bold]: {percent}")
            detail_list.append(f"{io_type.capitalize()} - {count:,} ops ({percent})")

        return RuleResult(
            compact_desc=' - '.join(compact_desc),
            description='I/O Operations',
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=total_count,
            value_fmt=f"{total_count:,} ops",
        )


class CharacteristicIOSizeRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.IO_SIZE.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        tasks = {}
        tasks['total_size'] = main_view['data_size'].sum()
        for io_type in IO_TYPES:
            if io_type != 'metadata':
                size_col = f"{io_type}_size"
                tasks[size_col] = main_view[size_col].sum()
        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        total_size = int(result['total_size'])

        value_fmt = format_bytes(total_size)

        compact_desc = [value_fmt]
        detail_list = []

        if total_size > 0:
            for i, io_type in enumerate(IO_TYPES):
                if io_type != 'metadata':
                    size_col = f"{io_type}_size"
                    size = int(result[size_col])
                    compact_desc.append((f"[bold]{COMPACT_IO_TYPES[i]}[/bold]: {size / total_size * 100:.2f}%"))
                    detail_list.append(
                        (f"{io_type.capitalize()} - {format_bytes(size)} ({size / total_size * 100:.2f}%)")
                    )

        return RuleResult(
            compact_desc=' - '.join(compact_desc),
            description='I/O Size',
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=total_size,
            value_fmt=value_fmt,
        )


class CharacteristicIOTimeRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.IO_TIME.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        view_types = main_view.index._meta.names

        tasks = {}

        if COL_PROC_NAME in view_types:
            tasks['total_time'] = main_view.groupby(['proc_name']).sum()['time'].max()

            for io_type in IO_TYPES:
                time_col = f"{io_type}_time"
                tasks[time_col] = main_view.groupby(['proc_name']).sum()[time_col].max()

        else:
            tasks['total_time'] = main_view['time'].sum()

            for io_type in IO_TYPES:
                time_col = f"{io_type}_time"
                tasks[time_col] = main_view[time_col].sum()

        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        total_time = result['total_time']

        compact_desc = [f"{total_time:.2f} s"]
        detail_list = []

        for i, io_type in enumerate(IO_TYPES):
            time_col = f"{io_type}_time"
            time = result[time_col]
            time_per = f"{time / total_time * 100:.2f}%"
            compact_desc.append(f"[bold]{COMPACT_IO_TYPES[i]}[/bold]: {time_per}")
            detail_list.append(f"{io_type.capitalize()} - {time:.2f} seconds ({time_per})")

        return RuleResult(
            compact_desc=' - '.join(compact_desc),
            description='I/O Time',
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            # rule=rule,
            value=total_time,
            value_fmt=f"{total_time:.2f} seconds",
        )


class CharacteristicProcessCount(CharacteristicRule):
    def __init__(self, rule_key: str) -> None:
        super().__init__(rule_key=rule_key)
        self.col = COL_PROC_NAME
        self.description = 'Processes/Ranks'
        if rule_key is KnownCharacteristics.APP_COUNT.value:
            self.col = COL_APP_NAME
            self.description = 'Apps'
            self.deps = [
                KnownCharacteristics.IO_COUNT.value,
                KnownCharacteristics.IO_SIZE.value,
                KnownCharacteristics.IO_TIME.value,
            ]
        elif rule_key is KnownCharacteristics.NODE_COUNT.value:
            self.col = COL_NODE_NAME
            self.description = 'Nodes'
            self.deps = [
                KnownCharacteristics.IO_COUNT.value,
                KnownCharacteristics.IO_SIZE.value,
                KnownCharacteristics.IO_TIME.value,
            ]

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        view_types = main_view.index._meta.names

        tasks = {}

        if COL_PROC_NAME not in view_types:
            if self.col == COL_PROC_NAME:
                tasks[f"{self.col}s"] = 0
                return tasks
            else:
                tasks[f"{self.col}s"] = pd.DataFrame()
                return tasks

        if self.col == COL_PROC_NAME:
            tasks[f"{self.col}s"] = main_view.map_partitions(
                lambda df: df.index.get_level_values(COL_PROC_NAME)
            ).nunique()
        else:
            tasks[f"{self.col}s"] = (
                main_view.map_partitions(set_proc_name_parts)
                .reset_index()
                .groupby([self.col, COL_PROC_NAME])
                .agg(
                    {
                        'count': sum,
                        'time': sum,
                        'read_size': sum,
                        'write_size': sum,
                    }
                )
                .groupby([self.col])
                .agg(
                    {
                        'count': sum,
                        'time': max,
                        'read_size': sum,
                        'write_size': sum,
                    }
                )
                .sort_values('time', ascending=False)
            )

        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        if self.col == COL_PROC_NAME:
            num_processes = int(result[f"{self.col}s"])

            value_fmt = f"{num_processes:,} {self.pluralize.plural_noun('process', num_processes)}"

            return RuleResult(
                compact_desc=value_fmt,
                description=self.description,
                detail_list=None,
                extra_data=None,
                object_hash=None,
                reasons=None,
                value=num_processes,
                value_fmt=value_fmt,
            )

        max_io_time = characteristics[KnownCharacteristics.IO_TIME.value].value
        total_ops = characteristics[KnownCharacteristics.IO_COUNT.value].value
        total_size = characteristics[KnownCharacteristics.IO_SIZE.value].value

        nodes_apps = pd.DataFrame(result[f"{self.col}s"])

        detail_list = []
        if len(nodes_apps) > 1 and total_size > 0:
            for node, row in nodes_apps.iterrows():
                read_size = row['read_size']
                write_size = row['write_size']
                read_size_fmt = format_bytes(read_size)
                write_size_fmt = format_bytes(write_size)
                read_size_per = read_size / total_size * 100
                write_size_per = write_size / total_size * 100
                detail_list.append(
                    ' - '.join(
                        [
                            node,
                            f"{row['time']:.2f} s ({row['time'] / max_io_time * 100:.2f}%)",
                            f"{read_size_fmt}/{write_size_fmt} R/W ({read_size_per:.2f}/{write_size_per:.2f}%)",
                            f"{int(row['count']):,} ops ({row['count'] / total_ops * 100:.2f}%)",
                        ]
                    )
                )

        num_nodes_apps = len(nodes_apps)

        if self.col == COL_NODE_NAME:
            value_fmt = f"{num_nodes_apps} {self.pluralize.plural_noun('node', num_nodes_apps)}"
        else:
            value_fmt = f"{num_nodes_apps} {self.pluralize.plural_noun('app', num_nodes_apps)}"

        return RuleResult(
            compact_desc=value_fmt,
            description=self.description,
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=num_nodes_apps,
            value_fmt=value_fmt,
        )


class CharacteristicTimePeriodCountRule(CharacteristicRule):
    def __init__(self) -> None:
        super().__init__(rule_key=KnownCharacteristics.TIME_PERIOD.value)

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        x = main_view.reset_index()

        tasks = {}

        if COL_TIME_RANGE in x.columns:
            tasks['total_count'] = x[COL_TIME_RANGE].nunique()
        else:
            tasks['total_count'] = 0

        return tasks

    def handle_task_results(
        self,
        result: Dict[str, int],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        num_time_periods = int(result["total_count"])
        compact_desc = f"{num_time_periods:,} {self.pluralize.plural_noun('time period', num_time_periods)}"
        time_granularity = raw_stats['time_granularity'] if isinstance(raw_stats, dict) else raw_stats.time_granularity
        return RuleResult(
            compact_desc=compact_desc,
            description='Time Periods',
            detail_list=None,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=num_time_periods,
            value_fmt=f"{compact_desc} (Time Granularity: {float(time_granularity):,})",
        )


class CharacteristicXferSizeRule(CharacteristicRule):
    def __init__(self, rule_key: str) -> None:
        super().__init__(rule_key)
        self.io_op = 'write' if rule_key is KnownCharacteristics.WRITE_XFER_SIZE.value else 'read'

    def define_tasks(
        self,
        main_view: dd.DataFrame,
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
    ) -> Dict[str, Delayed]:
        tasks = {}

        count_col, min_col, max_col, per_col, xfer_col = (
            f"{self.io_op}_count",
            f"{self.io_op}_min",
            f"{self.io_op}_max",
            'per',
            'xfer',
        )

        min_view = main_view[main_view[min_col] > 0]
        max_view = main_view[main_view[max_col] > 0]

        tasks['min_xfer_size'] = min_view[min_col].min()
        tasks['max_xfer_size'] = max_view[max_col].max()

        tasks['xfer_sizes'] = max_view.groupby(max_col)[count_col].sum()

        return tasks

    def handle_task_results(
        self,
        result: Dict[str, Union[str, int, pd.DataFrame, pd.Series]],
        characteristics: Dict[str, RuleResult] = None,
        raw_stats: RawStats = None,
    ) -> RuleResult:
        count_col, min_col, max_col, per_col, xfer_col = (
            f"{self.io_op}_count",
            f"{self.io_op}_min",
            f"{self.io_op}_max",
            'per',
            'xfer',
        )

        min_xfer_size = 0
        max_xfer_size = 0
        if not np.isnan(result['min_xfer_size']):
            min_xfer_size = int(result['min_xfer_size'])
        if not np.isnan(result['max_xfer_size']):
            max_xfer_size = int(result['max_xfer_size'])

        xfer_sizes = pd.DataFrame(result['xfer_sizes'])
        xfer_sizes[xfer_col] = pd.cut(
            xfer_sizes.index,
            bins=XFER_SIZE_BINS,
            labels=XFER_SIZE_BIN_LABELS,
            right=True,
        )
        xfer_bins = xfer_sizes.groupby([xfer_col], observed=True).sum().replace(0, np.nan).dropna()
        xfer_bins.loc[:, per_col] = xfer_bins[count_col] / xfer_bins[count_col].sum()

        total_ops = int(xfer_bins[count_col].sum())

        compact_desc = self._get_xfer_size(max_xfer_size)

        if min_xfer_size > 0 and max_xfer_size > 0:
            if min_xfer_size == max_xfer_size:
                compact_desc = f"{self._get_xfer_size(min_xfer_size, True)}-{self._get_xfer_size(max_xfer_size)}"
            else:
                compact_desc = f"{self._get_xfer_size(min_xfer_size)}-{self._get_xfer_size(max_xfer_size)}"

        detail_list = []
        for xfer, row in xfer_bins.iterrows():
            detail_list.append(f"{xfer} - {int(row[count_col]):,} ops ({row['per'] * 100:.2f}%)")

        result = RuleResult(
            compact_desc=compact_desc,
            description='Write Requests' if self.io_op == 'write' else 'Read Requests',
            detail_list=detail_list,
            extra_data=None,
            object_hash=None,
            reasons=None,
            value=(min_xfer_size, max_xfer_size),
            value_fmt=f"{compact_desc} - {total_ops:,} ops",
        )

        return result

    @staticmethod
    def _get_xfer_size(size: float, previous=False):
        size_bin = np.digitize(size, bins=XFER_SIZE_BINS, right=True)
        if previous:
            size_bin = size_bin - 1
        size_label = np.choose(size_bin, choices=XFER_SIZE_BIN_NAMES, mode='clip')
        return size_label



import numpy as np
from json import JSONEncoder


class NpEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return super(NpEncoder, self).default(obj)



import logging
import socket
from dataclasses import asdict, dataclass, field
from hydra.core.config_store import ConfigStore
from hydra.conf import HelpConf, JobConf
from omegaconf import MISSING
from typing import Any, Dict, List, Optional

from .constants import VIEW_TYPES


@dataclass
class AnalyzerConfig:
    bottleneck_dir: Optional[str] = "${hydra:runtime.output_dir}/bottlenecks"
    checkpoint: Optional[bool] = True
    checkpoint_dir: Optional[str] = "${hydra:runtime.output_dir}/checkpoints"
    time_approximate: Optional[bool] = True
    time_granularity: Optional[float] = MISSING


@dataclass
class DarshanAnalyzerConfig(AnalyzerConfig):
    _target_: str = "wisio.darshan.DarshanAnalyzer"
    time_granularity: Optional[float] = 1e3


@dataclass
class DFTracerAnalyzerConfig(AnalyzerConfig):
    _target_: str = "wisio.dftracer.DFTracerAnalyzer"
    time_granularity: Optional[float] = 1e6


@dataclass
class RecorderAnalyzerConfig(AnalyzerConfig):
    _target_: str = "wisio.recorder.RecorderAnalyzer"
    time_granularity: Optional[float] = 1e7


@dataclass
class ClusterConfig:
    local_directory: Optional[str] = "/tmp/${hydra:job.name}/${hydra:job.id}"


@dataclass
class JobQueueClusterSchedulerConfig:
    dashboard_address: Optional[str] = None
    host: Optional[str] = field(default_factory=socket.gethostname)


@dataclass
class JobQueueClusterConfig(ClusterConfig):
    cores: int = 16
    death_timeout: Optional[int] = 60
    job_directives_skip: Optional[List[str]] = field(default_factory=list)
    job_extra_directives: Optional[List[str]] = field(default_factory=list)
    log_directory: Optional[str] = ""
    memory: Optional[str] = None
    processes: Optional[int] = 1
    scheduler_options: Optional[JobQueueClusterSchedulerConfig] = field(
        default_factory=JobQueueClusterSchedulerConfig
    )


@dataclass
class LocalClusterConfig(ClusterConfig):
    _target_: str = "dask.distributed.LocalCluster"
    host: Optional[str] = None
    memory_limit: Optional[int] = None
    n_workers: Optional[int] = None
    processes: Optional[bool] = True
    silence_logs: Optional[int] = logging.CRITICAL


@dataclass
class LSFClusterConfig(JobQueueClusterConfig):
    _target_: str = "dask_jobqueue.LSFCluster"
    use_stdin: Optional[bool] = True


@dataclass
class PBSClusterConfig(JobQueueClusterConfig):
    _target_: str = "dask_jobqueue.PBSCluster"


@dataclass
class SLURMClusterConfig(JobQueueClusterConfig):
    _target_: str = "dask_jobqueue.SLURMCluster"


@dataclass
class OutputConfig:
    compact: Optional[bool] = False
    group_behavior: Optional[bool] = False
    name: Optional[str] = ""
    root_only: Optional[bool] = True
    view_names: Optional[List[str]] = field(default_factory=list)


@dataclass
class ConsoleOutputConfig(OutputConfig):
    _target_: str = "wisio.output.ConsoleOutput"
    max_bottlenecks: Optional[int] = 3
    show_debug: Optional[bool] = False
    show_characteristics: Optional[bool] = True
    show_header: Optional[bool] = True


@dataclass
class CSVOutputConfig(OutputConfig):
    _target_: str = "wisio.output.CSVOutput"


@dataclass
class SQLiteOutputConfig(OutputConfig):
    _target_: str = "wisio.output.SQLiteOutput"
    run_db_path: Optional[str] = ""


@dataclass
class CustomJobConfig(JobConf):
    name: str = "wisio"


@dataclass
class CustomHelpConfig(HelpConf):
    app_name: str = "WisIO"
    header: str = "${hydra:help.app_name}: Workflow I/O Analysis Tool"
    footer: str = field(
        default_factory=lambda: """
Powered by Hydra (https://hydra.cc)

Use --hydra-help to view Hydra specific help
    """.strip()
    )
    template: str = field(
        default_factory=lambda: """
${hydra:help.header}

== Configuration groups ==

Compose your configuration from those groups (group=option)

$APP_CONFIG_GROUPS
== Config ==

Override anything in the config (foo.bar=value)

$CONFIG
${hydra:help.footer}
    """.strip()
    )


@dataclass
class CustomLoggingConfig:
    version: int = 1
    formatters: Dict[str, Any] = field(
        default_factory=lambda: {
            "simple": {
                "datefmt": "%H:%M:%S",
                "format": "[%(levelname)s] [%(asctime)s.%(msecs)03d] %(message)s [%(pathname)s:%(lineno)d]",
            }
        }
    )
    handlers: Dict[str, Any] = field(
        default_factory=lambda: {
            "file": {
                "class": "logging.FileHandler",
                "formatter": "simple",
                "filename": "${hydra:runtime.output_dir}/${hydra:job.name}.log",
            },
        }
    )
    root: Dict[str, Any] = field(
        default_factory=lambda: {
            "level": "INFO",
            "handlers": ["file"],
        }
    )
    disable_existing_loggers: bool = False


@dataclass
class Config:
    defaults: List[Any] = field(
        default_factory=lambda: [
            {"hydra/job": "custom"},
            {"cluster": "local"},
            {"output": "console"},
            "_self_",
            {"override hydra/help": "custom"},
            {"override hydra/job_logging": "custom"},
        ]
    )
    analyzer: AnalyzerConfig = MISSING
    cluster: ClusterConfig = MISSING
    debug: Optional[bool] = False
    exclude_bottlenecks: Optional[List[str]] = field(default_factory=list)
    exclude_characteristics: Optional[List[str]] = field(default_factory=list)
    metrics: Optional[List[str]] = field(default_factory=lambda: ["iops"])
    logical_view_types: Optional[bool] = False
    output: OutputConfig = MISSING
    percentile: Optional[float] = None
    threshold: Optional[int] = None
    time_granularity: Optional[float] = 1e6
    trace_path: str = MISSING
    verbose: Optional[bool] = False
    view_types: Optional[List[str]] = field(default_factory=lambda: VIEW_TYPES)


def init_hydra_config_store() -> None:
    cs = ConfigStore.instance()
    cs.store(group="hydra/help", name="custom", node=asdict(CustomHelpConfig()))
    cs.store(group="hydra/job", name="custom", node=CustomJobConfig)
    cs.store(group="hydra/job_logging", name="custom", node=CustomLoggingConfig)
    cs.store(name="config", node=Config)
    cs.store(group="analyzer", name="darshan", node=DarshanAnalyzerConfig)
    cs.store(group="analyzer", name="dftracer", node=DFTracerAnalyzerConfig)
    cs.store(group="analyzer", name="recorder", node=RecorderAnalyzerConfig)
    cs.store(group="cluster", name="local", node=LocalClusterConfig)
    cs.store(group="cluster", name="lsf", node=LSFClusterConfig)
    cs.store(group="cluster", name="pbs", node=PBSClusterConfig)
    cs.store(group="cluster", name="slurm", node=SLURMClusterConfig)
    cs.store(group="output", name="console", node=ConsoleOutputConfig)
    cs.store(group="output", name="csv", node=CSVOutputConfig)
    cs.store(group="output", name="sqlite", node=SQLiteOutputConfig)



import logging
from time import perf_counter


class ElapsedTimeLogger(object):
    def __init__(self, message: str, level=logging.INFO, stacklevel=3):
        self.level = level
        self.message = message
        self.stacklevel = stacklevel

    def __enter__(self):
        self.start_time = perf_counter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end_time = perf_counter()
        self.elapsed_time = self.end_time - self.start_time
        if self.level is logging.DEBUG:
            logging.debug(
                msg=f"{self.message} ({self.elapsed_time})",
                stacklevel=self.stacklevel,
            )
        else:
            logging.info(
                msg=f"{self.message} ({self.elapsed_time})",
                stacklevel=self.stacklevel,
            )



import hydra
from dask_jobqueue import LSFCluster, PBSCluster, SLURMCluster
from distributed import Client, LocalCluster
from hydra.utils import instantiate
from typing import Union

from .analyzer import Analyzer
from .config import Config, init_hydra_config_store
from .dftracer import DFTracerAnalyzer
from .output import ConsoleOutput, CSVOutput, SQLiteOutput
from .recorder import RecorderAnalyzer


try:
    from .darshan import DarshanAnalyzer
except ModuleNotFoundError:
    DarshanAnalyzer = Analyzer


AnalyzerType = Union[DarshanAnalyzer, DFTracerAnalyzer, RecorderAnalyzer]
ClusterType = Union[LocalCluster, LSFCluster, PBSCluster, SLURMCluster]
OutputType = Union[ConsoleOutput, CSVOutput, SQLiteOutput]


init_hydra_config_store()


@hydra.main(version_base=None, config_name="config")
def main(cfg: Config) -> None:
    cluster: ClusterType = instantiate(cfg.cluster)
    client = Client(cluster)
    analyzer: AnalyzerType = instantiate(
        cfg.analyzer,
        debug=cfg.debug,
        verbose=cfg.verbose,
    )
    result = analyzer.analyze_trace(
        trace_path=cfg.trace_path,
        # accuracy=cfg.accuracy,
        exclude_bottlenecks=cfg.exclude_bottlenecks,
        exclude_characteristics=cfg.exclude_characteristics,
        logical_view_types=cfg.logical_view_types,
        metrics=cfg.metrics,
        percentile=cfg.percentile,
        threshold=cfg.threshold,
        view_types=cfg.view_types,
    )
    output: OutputType = instantiate(cfg.output)
    output.handle_result(metrics=cfg.metrics, result=result)


if __name__ == "__main__":
    main()



from decimal import Decimal


def _round_num(n, decimal=2):
    n = Decimal(n)
    return n.to_integral() if n == n.to_integral() else round(n.normalize(), decimal)


def format_number(value):
    return format(int(value), ',d')


def numerize(n, decimal=2):
    if n == 0:
        return n
    suffixes = ["", "K", "M", "B", "T", "Qa", "Qu", "S", "Oc", "No",
                "D", "Ud", "Dd", "Td", "Qt", "Qi", "Se", "Od", "Nd", "V",
                "Uv", "Dv", "Tv", "Qv", "Qx", "Sx", "Ox", "Nx", "Tn", "Qa",
                "Qu", "S", "Oc", "No", "D", "Ud", "Dd", "Td", "Qt", "Qi",
                "Se", "Od", "Nd", "V", "Uv", "Dv", "Tv", "Qv", "Qx", "Sx",
                "Ox", "Nx", "Tn", "x", "xx", "xxx", "X", "XX", "XXX", "END"]

    sci_expr = [1e0, 1e3, 1e6, 1e9, 1e12, 1e15, 1e18, 1e21, 1e24, 1e27,
                1e30, 1e33, 1e36, 1e39, 1e42, 1e45, 1e48, 1e51, 1e54, 1e57,
                1e60, 1e63, 1e66, 1e69, 1e72, 1e75, 1e78, 1e81, 1e84, 1e87,
                1e90, 1e93, 1e96, 1e99, 1e102, 1e105, 1e108, 1e111, 1e114, 1e117,
                1e120, 1e123, 1e126, 1e129, 1e132, 1e135, 1e138, 1e141, 1e144, 1e147,
                1e150, 1e153, 1e156, 1e159, 1e162, 1e165, 1e168, 1e171, 1e174, 1e177]
    minus_buff = n
    n = abs(n)
    for x in range(len(sci_expr)):
        try:
            if n >= sci_expr[x] and n < sci_expr[x+1]:
                suffix = suffixes[x]
                if n >= 1e3:
                    num = str(_round_num(n/sci_expr[x], decimal))
                else:
                    num = str(n)
                return num + suffix if minus_buff > 0 else "-" + num + suffix
        except IndexError:
            print("You've reached the end")



import dask.dataframe as dd
import json
import numpy as np
import pandas as pd
from dask.distributed import Future, get_client
from typing import Union

from .analyzer import Analyzer
from .constants import IO_CATS


CAT_POSIX = 0
DROPPED_COLS = [
    'app',
    'bandwidth',
    'file_id',
    'hostname',
    'index',
    'level',
    'proc',
    'proc_id',
    'rank',
    'tend',
    'thread_id',
    'tmid',
    'tstart',
]
RENAMED_COLS = {'duration': 'time'}


class RecorderAnalyzer(Analyzer):
    def read_trace(self, trace_path: str) -> dd.DataFrame:
        self.global_min_max = self._load_global_min_max(trace_path=trace_path)
        return dd.read_parquet(trace_path)

    def postread_trace(self, traces: dd.DataFrame) -> dd.DataFrame:
        traces['acc_pat'] = traces['acc_pat'].astype(np.uint8)
        traces['count'] = 1
        traces['duration'] = traces['duration'].astype(np.float64)
        traces['io_cat'] = traces['io_cat'].astype(np.uint8)
        time_ranges = self._compute_time_ranges(
            global_min_max=self.global_min_max,
            time_granularity=self.time_granularity,
        )
        traces = (
            traces[(traces['cat'] == CAT_POSIX) & (traces['io_cat'].isin(IO_CATS))]
            .map_partitions(self._set_time_ranges, time_ranges=time_ranges)
            .rename(columns=RENAMED_COLS)
            .drop(columns=DROPPED_COLS, errors='ignore')
        )
        return traces

    def compute_total_count(self, traces: dd.DataFrame) -> int:
        return (
            traces[(traces['cat'] == CAT_POSIX) & (traces['io_cat'].isin(IO_CATS))]
            .index.count()
            .persist()
        )

    @staticmethod
    def _compute_time_ranges(global_min_max: dict, time_granularity: int):
        tmid_min, tmid_max = global_min_max['tmid']
        time_ranges = np.arange(tmid_min, tmid_max, time_granularity)
        return get_client().scatter(time_ranges)

    @staticmethod
    def _load_global_min_max(trace_path: str) -> dict:
        with open(f"{trace_path}/global.json") as file:
            global_min_max = json.load(file)
        return global_min_max

    @staticmethod
    def _set_time_ranges(df: pd.DataFrame, time_ranges: Union[Future, np.ndarray]):
        if isinstance(time_ranges, Future):
            time_ranges = time_ranges.result()
        return df.assign(
            time_range=np.digitize(df['tmid'], bins=time_ranges, right=True)
        )



import dask
import dask.bag as db
import dask.dataframe as dd
import portion as P
import math
import json
import logging
import os
import sys
import zindex_py as zindex
from dask.distributed import wait
from glob import glob

from .analyzer import Analyzer
from .constants import (
    COL_ACC_PAT,
    COL_COUNT,
    COL_FILE_NAME,
    COL_FUNC_ID,
    COL_HOST_NAME,
    COL_IO_CAT,
    COL_PROC_NAME,
    COL_TIME,
    COL_TIME_RANGE,
    IOCategory,
)


CAT_POSIX = 'POSIX'
DFTRACER_TIME_RESOLUTION = 1e6
PFW_COL_MAPPING = {
    'name': COL_FUNC_ID,
    'dur': COL_TIME,
    'hhash': COL_HOST_NAME,
    'fhash': COL_FILE_NAME,
    'trange': COL_TIME_RANGE,
}


def create_index(filename):
    index_file = f"{filename}.zindex"
    if not os.path.exists(index_file):
        status = zindex.create_index(
            filename,
            index_file=f"file:{index_file}",
            regex="id:\b([0-9]+)",
            numeric=True,
            unique=True,
            debug=False,
            verbose=False,
        )
        logging.debug(f"Creating Index for {filename} returned {status}")
    return filename


def generate_line_batches(filename, max_line):
    batch_size = 1024 * 16
    for start in range(0, max_line, batch_size):
        end = min((start + batch_size - 1), (max_line - 1))
        logging.debug(f"Created a batch for {filename} from [{start}, {end}] lines")
        yield filename, start, end


def get_linenumber(filename):
    index_file = f"{filename}.zindex"
    line_number = zindex.get_max_line(
        filename,
        index_file=index_file,
        debug=False,
        verbose=False,
    )
    logging.debug(f" The {filename} has {line_number} lines")
    return (filename, line_number)


def get_size(filename):
    if filename.endswith('.pfw'):
        size = os.stat(filename).st_size
    elif filename.endswith('.pfw.gz'):
        index_file = f"{filename}.zindex"
        line_number = zindex.get_max_line(
            filename,
            index_file=index_file,
            debug=False,
            verbose=False,
        )
        size = line_number * 256
    logging.debug(f" The {filename} has {size / 1024**3} GB size")
    return int(size)


def get_conditions_default(json_obj):
    io_cond = "POSIX" == json_obj["cat"]
    return False, False, io_cond


def io_columns(time_approximate=True):
    if is_pyarrow_dtype_supported():
        cols = {
            'compute_time': 'string[pyarrow]',
            'io_time': 'string[pyarrow]',
            'app_io_time': 'string[pyarrow]',
            'total_time': 'string[pyarrow]',
            'fhash': 'uint64[pyarrow]',
            'hhash': 'uint64[pyarrow]',
            'phase': 'uint16[pyarrow]',
            'size': 'uint64[pyarrow]',
        }
        if time_approximate:
            cols['compute_time'] = 'uint64[pyarrow]'
            cols['io_time'] = 'uint64[pyarrow]'
            cols['app_io_time'] = 'uint64[pyarrow]'
            cols['total_time'] = 'uint64[pyarrow]'
        return cols
    return {
        'compute_time': 'Int64' if time_approximate else 'string',
        'io_time': 'Int64' if time_approximate else 'string',
        'app_io_time': 'Int64' if time_approximate else 'string',
        'total_time': 'Int64' if time_approximate else 'string',
        'fhash': 'Int64',
        'hhash': 'Int64',
        'phase': 'Int16',
        'size': 'Int64',
    }


def io_function(json_object, current_dict, time_approximate, condition_fn):
    d = {}
    d["phase"] = 0
    if not condition_fn:
        condition_fn = get_conditions_default
    app_io_cond, compute_cond, io_cond = condition_fn(json_object)
    if time_approximate:
        d["total_time"] = 0
        if compute_cond:
            d["compute_time"] = current_dict["dur"]
            d["total_time"] = current_dict["dur"]
            d["phase"] = 1
        elif io_cond:
            d["io_time"] = current_dict["dur"]
            d["total_time"] = current_dict["dur"]
            d["phase"] = 2
        elif app_io_cond:
            d["total_time"] = current_dict["dur"]
            d["app_io_time"] = current_dict["dur"]
            d["phase"] = 3
    else:
        if compute_cond:
            d["compute_time"] = current_dict["tinterval"]
            d["total_time"] = current_dict["tinterval"]
            d["phase"] = 1
        elif io_cond:
            d["io_time"] = current_dict["tinterval"]
            d["total_time"] = current_dict["tinterval"]
            d["phase"] = 2
        elif app_io_cond:
            d["app_io_time"] = current_dict["tinterval"]
            d["total_time"] = current_dict["tinterval"]
            d["phase"] = 3
        else:
            d["total_time"] = P.to_string(P.empty())
            d["io_time"] = P.to_string(P.empty())
    if "args" in json_object:
        if "fhash" in json_object["args"]:
            if type(json_object["args"]["fhash"]) is str:
                d["fhash"] = int(json_object["args"]["fhash"], 16)
            else:
                d["fhash"] = json_object["args"]["fhash"]
        if "POSIX" == json_object["cat"] and "ret" in json_object["args"]:
            size = int(json_object["args"]["ret"])
            if size > 0:
                if "write" in json_object["name"]:
                    d["size"] = size
                elif (
                    "read" in json_object["name"]
                    and "readdir" not in json_object["name"]
                ):
                    d["size"] = size
        else:
            if "image_size" in json_object["args"]:
                size = int(json_object["args"]["image_size"])
                if size > 0:
                    d["size"] = size
    return d


def is_pyarrow_dtype_supported() -> bool:
    return sys.version_info >= (3, 9)


def load_indexed_gzip_files(filename, start, end):
    index_file = f"{filename}.zindex"
    json_lines = zindex.zquery(
        filename,
        index_file=index_file,
        raw=f"select a.line from LineOffsets a where a.line >= {start} AND a.line <= {end};",
        debug=False,
        verbose=False,
    )
    logging.debug(f"Read {len(json_lines)} json lines for [{start}, {end}]")
    return json_lines


def load_objects(line, fn, time_granularity, time_approximate, condition_fn, load_data):
    d = {}
    if (
        line is not None
        and line != ""
        and len(line) > 0
        and "[" != line[0]
        and "]" != line[0]
        and line != "\n"
    ):
        val = {}
        try:
            unicode_line = ''.join([i if ord(i) < 128 else '#' for i in line])
            val = json.loads(unicode_line, strict=False)
            logging.debug(f"Loading dict {val}")
            if "name" in val:
                d["name"] = val["name"]
            if "cat" in val:
                d["cat"] = val["cat"]
            if "pid" in val:
                d["pid"] = val["pid"]
            if "tid" in val:
                d["tid"] = val["tid"]
            if "args" in val:
                if "hhash" in val["args"]:
                    if type(val["args"]["hhash"]) is str:
                        d["hhash"] = int(val["args"]["hhash"], 16)
                    else:
                        d["hhash"] = val["args"]["hhash"]
                if "level" in val["args"]:
                    d["level"] = int(val["args"]["level"])
            if "M" == val["ph"]:
                if d["name"] == "FH":
                    d["type"] = 1  # 1-> file hash
                    if (
                        "args" in val
                        and "name" in val["args"]
                        and "value" in val["args"]
                    ):
                        d["name"] = val["args"]["name"]
                        if type(val["args"]["value"]) is str:
                            d["hash"] = int(val["args"]["value"], 16)
                        else:
                            d["hash"] = val["args"]["value"]
                            # TODO(izzet): maybe add hash here
                elif d["name"] == "HH":
                    d["type"] = 2  # 2-> hostname hash
                    if (
                        "args" in val
                        and "name" in val["args"]
                        and "value" in val["args"]
                    ):
                        d["name"] = val["args"]["name"]
                        if type(val["args"]["value"]) is str:
                            d["hash"] = int(val["args"]["value"], 16)
                        else:
                            d["hash"] = val["args"]["value"]
                elif d["name"] == "SH":
                    d["type"] = 3  # 3-> string hash
                    if (
                        "args" in val
                        and "name" in val["args"]
                        and "value" in val["args"]
                    ):
                        d["name"] = val["args"]["name"]
                        if type(val["args"]["value"]) is str:
                            d["hash"] = int(val["args"]["value"], 16)
                        else:
                            d["hash"] = val["args"]["value"]
                elif d["name"] == "PR":
                    d["type"] = 5  # 5-> process metadata
                    if (
                        "args" in val
                        and "name" in val["args"]
                        and "value" in val["args"]
                    ):
                        d["name"] = val["args"]["name"]
                        if type(val["args"]["value"]) is str:
                            d["hash"] = int(val["args"]["value"], 16)
                        else:
                            d["hash"] = val["args"]["value"]
                else:
                    d["type"] = 4  # 4-> others
                    if (
                        "args" in val
                        and "name" in val["args"]
                        and "value" in val["args"]
                    ):
                        d["name"] = val["args"]["name"]
                        d["value"] = str(val["args"]["value"])
            else:
                d["type"] = 0  # 0->regular event
                if "dur" in val:
                    val["dur"] = int(val["dur"])
                    val["ts"] = int(val["ts"])
                    d["ts"] = val["ts"]
                    d["dur"] = val["dur"]
                    d["te"] = d["ts"] + d["dur"]
                    if not time_approximate:
                        d["tinterval"] = P.to_string(
                            P.closed(val["ts"], val["ts"] + val["dur"])
                        )
                    d["trange"] = int(
                        ((val["ts"] + val["dur"]) / 2.0) / time_granularity
                    )
                d.update(io_function(val, d, time_approximate, condition_fn))
            logging.debug(f"built an dictionary for line {d}")
            yield d
        except ValueError as error:
            logging.error(f"Processing {line} failed with {error}")
    return {}


class DFTracerAnalyzer(Analyzer):
    def read_trace(self, trace_path: str) -> dd.DataFrame:
        conditions = None
        load_cols = {}
        load_data = {}
        load_fn = None
        metadata_cols = {}
        if os.path.isdir(trace_path) and '*' not in trace_path:
            trace_path = f"{trace_path}/*.pfw*"
        # ===============================================
        file_pattern = glob(trace_path)
        all_files = []
        pfw_pattern = []
        pfw_gz_pattern = []
        for file in file_pattern:
            if file.endswith('.pfw'):
                pfw_pattern.append(file)
                all_files.append(file)
            elif file.endswith('.pfw.gz'):
                pfw_gz_pattern.append(file)
                all_files.append(file)
            else:
                logging.warning(f"Ignoring unsuported file {file}")
        if len(all_files) == 0:
            logging.error(f"No files selected for .pfw and .pfw.gz")
            exit(1)
        logging.debug(f"Processing files {all_files}")
        delayed_indices = []
        if len(pfw_gz_pattern) > 0:
            db.from_sequence(pfw_gz_pattern).map(create_index).compute()
        logging.info(f"Created index for {len(pfw_gz_pattern)} files")
        total_size = db.from_sequence(all_files).map(get_size).sum().compute()
        logging.info(f"Total size of all files are {total_size} bytes")
        gz_bag = None
        pfw_bag = None
        if len(pfw_gz_pattern) > 0:
            max_line_numbers = (
                db.from_sequence(pfw_gz_pattern).map(get_linenumber).compute()
            )
            logging.debug(f"Max lines per file are {max_line_numbers}")
            json_line_delayed = []
            total_lines = 0
            for filename, max_line in max_line_numbers:
                total_lines += max_line
                for _, start, end in generate_line_batches(filename, max_line):
                    json_line_delayed.append((filename, start, end))

            logging.info(
                f"Loading {len(json_line_delayed)} batches out of {len(pfw_gz_pattern)} files and has {total_lines} lines overall"
            )
            json_line_bags = []
            for filename, start, end in json_line_delayed:
                num_lines = end - start + 1
                json_line_bags.append(
                    dask.delayed(load_indexed_gzip_files, nout=num_lines)(
                        filename, start, end
                    )
                )
            json_lines = db.concat(json_line_bags)
            gz_bag = (
                json_lines.map(
                    load_objects,
                    fn=load_fn,
                    time_granularity=self.time_granularity,
                    time_approximate=self.time_approximate,
                    condition_fn=conditions,
                    load_data=load_data,
                )
                .flatten()
                .filter(lambda x: "name" in x)
            )
        main_bag = None
        if len(pfw_pattern) > 0:
            pfw_bag = (
                db.read_text(pfw_pattern)
                .map(
                    load_objects,
                    fn=load_fn,
                    time_granularity=self.time_granularity,
                    time_approximate=self.time_approximate,
                    condition_fn=conditions,
                    load_data=load_data,
                )
                .flatten()
                .filter(lambda x: "name" in x)
            )
        if len(pfw_gz_pattern) > 0 and len(pfw_pattern) > 0:
            main_bag = db.concat([pfw_bag, gz_bag])
        elif len(pfw_gz_pattern) > 0:
            main_bag = gz_bag
        elif len(pfw_pattern) > 0:
            main_bag = pfw_bag
        if main_bag:
            columns = {
                'name': 'string',
                'cat': 'string',
                'type': 'Int8',
                'pid': 'Int64',
                'tid': 'Int64',
                'ts': 'Int64',
                'te': 'Int64',
                'dur': 'Int64',
                'tinterval': 'Int64' if self.time_approximate else 'string',
                'trange': 'Int64',
                'level': 'Int8',
            }
            if is_pyarrow_dtype_supported():
                columns = {
                    'name': 'string[pyarrow]',
                    'cat': 'string[pyarrow]',
                    'type': 'uint8[pyarrow]',
                    'pid': 'uint64[pyarrow]',
                    'tid': 'uint64[pyarrow]',
                    'ts': 'uint64[pyarrow]',
                    'te': 'uint64[pyarrow]',
                    'dur': 'uint64[pyarrow]',
                    'tinterval': 'uint64[pyarrow]',
                    'trange': 'uint64[pyarrow]',
                    'level': 'uint8[pyarrow]',
                }
                if self.time_approximate:
                    columns['tinterval'] = 'string[pyarrow]'
            columns.update(io_columns())
            columns.update(load_cols)
            file_hash_columns = {
                'name': 'string',
                'hash': 'Int64',
                'pid': 'Int64',
                'tid': 'Int64',
                'hhash': 'Int64',
            }
            hostname_hash_columns = {
                'name': 'string',
                'hash': 'Int64',
                'pid': 'Int64',
                'tid': 'Int64',
                'hhash': 'Int64',
            }
            string_hash_columns = {
                'name': 'string',
                'hash': 'Int64',
                'pid': 'Int64',
                'tid': 'Int64',
                'hhash': 'Int64',
            }
            other_metadata_columns = {
                'name': 'string',
                'value': 'string',
                'pid': 'Int64',
                'tid': 'Int64',
                'hhash': 'Int64',
            }
            if is_pyarrow_dtype_supported():
                file_hash_columns = {
                    'name': 'string[pyarrow]',
                    'hash': 'uint64[pyarrow]',
                    'pid': 'uint64[pyarrow]',
                    'tid': 'uint64[pyarrow]',
                    'hhash': 'uint64[pyarrow]',
                }
                hostname_hash_columns = {
                    'name': 'string[pyarrow]',
                    'hash': 'uint64[pyarrow]',
                    'pid': 'uint64[pyarrow]',
                    'tid': 'uint64[pyarrow]',
                    'hhash': 'uint64[pyarrow]',
                }
                string_hash_columns = {
                    'name': 'string[pyarrow]',
                    'hash': 'uint64[pyarrow]',
                    'pid': 'uint64[pyarrow]',
                    'tid': 'uint64[pyarrow]',
                    'hhash': 'uint64[pyarrow]',
                }
                other_metadata_columns = {
                    'name': 'string[pyarrow]',
                    'value': 'string[pyarrow]',
                    'pid': 'uint64[pyarrow]',
                    'tid': 'uint64[pyarrow]',
                    'hhash': 'uint64[pyarrow]',
                }
            if "FH" in metadata_cols:
                file_hash_columns.update(metadata_cols["FH"])
            if "HH" in metadata_cols:
                hostname_hash_columns.update(metadata_cols["HH"])
            if "SH" in metadata_cols:
                string_hash_columns.update(metadata_cols["SH"])
            if "M" in metadata_cols:
                other_metadata_columns.update(metadata_cols["M"])
            columns.update(file_hash_columns)
            columns.update(hostname_hash_columns)
            columns.update(string_hash_columns)
            columns.update(other_metadata_columns)

            self.all_events = main_bag.to_dataframe(meta=columns)
            events = self.all_events.query("type == 0")
            self.file_hash = (
                self.all_events.query("type == 1")[list(file_hash_columns.keys())]
                .groupby('hash')
                .first()
                .persist()
            )
            self.host_hash = (
                self.all_events.query("type == 2")[list(hostname_hash_columns.keys())]
                .groupby('hash')
                .first()
                .persist()
            )
            self.string_hash = (
                self.all_events.query("type == 3")[list(string_hash_columns.keys())]
                .groupby('hash')
                .first()
                .persist()
            )
            self.metadata = self.all_events.query("type == 4")[
                list(other_metadata_columns.keys())
            ].persist()
            self.n_partition = math.ceil(total_size / (128 * 1024**2))
            logging.debug(f"Number of partitions used are {self.n_partition}")
            self.events = events.repartition(npartitions=self.n_partition).persist()
            _ = wait(self.events)
            self.events['ts'] = self.events['ts'] - self.events['ts'].min()
            self.events['te'] = self.events['ts'] + self.events['dur']
            self.events['trange'] = self.events['ts'] // self.time_granularity
            if is_pyarrow_dtype_supported():
                self.events['ts'] = self.events['ts'].astype('uint64[pyarrow]')
                self.events['te'] = self.events['te'].astype('uint64[pyarrow]')
                self.events['trange'] = self.events['trange'].astype('uint16[pyarrow]')
            else:
                self.events['ts'] = self.events['ts'].astype('Int64')
                self.events['te'] = self.events['te'].astype('Int64')
                self.events['trange'] = self.events['trange'].astype('Int16')
            self.events = self.events.persist()
            _ = wait(
                [
                    self.file_hash,
                    self.host_hash,
                    self.string_hash,
                    self.metadata,
                    self.events,
                ]
            )
        else:
            logging.error("Unable to load traces")
            exit(1)

        self.events['dur'] = self.events['dur'] / DFTRACER_TIME_RESOLUTION

        return self.events.rename(columns=PFW_COL_MAPPING)

    def postread_trace(self, traces: dd.DataFrame) -> dd.DataFrame:
        # 'name': COL_FUNC_ID,
        # 'dur': COL_TIME,
        # 'hostname': COL_HOST_NAME,
        # 'filename': COL_FILE_NAME,
        # traces[COL_FUNC_ID] = traces['name']
        # traces[COL_TIME] = traces['dur']
        # traces[COL_HOST_NAME] = traces['hostname']
        # traces[COL_FILE_NAME] = traces['filename']
        # traces = traces.rename(columns=PFW_COL_MAPPING)
        traces = traces[(traces['cat'] == CAT_POSIX) & (traces['ts'] > 0)]
        # traces[COL_TIME] = traces[COL_TIME] / DFTRACER_TIME_RESOLUTION
        # traces['ts'] = traces['ts'] - traces['ts'].min()
        # traces['ts'] = traces['ts'] / DFTRACER_TIME_RESOLUTION
        # traces['te'] = traces['ts'] + traces[COL_TIME]
        # traces[COL_TIME_RANGE] = (
        #     ((traces['te'] / self.time_granularity) * DFTRACER_TIME_RESOLUTION)
        #     .round()
        #     .astype(int)
        # )
        traces[COL_PROC_NAME] = (
            'app#'
            + traces[COL_HOST_NAME].astype(str)
            + '#'
            # + 'host#'
            + traces['pid'].astype(str)
            + '#'
            + traces['tid'].astype(str)
        )
        read_cond = 'read'
        write_cond = 'write'
        metadata_cond = 'readlink'
        traces[COL_ACC_PAT] = 0
        traces[COL_COUNT] = 1
        traces[COL_IO_CAT] = 0
        traces[COL_IO_CAT] = traces[COL_IO_CAT].mask(
            (traces['cat'] == CAT_POSIX)
            & ~traces[COL_FUNC_ID].str.contains(read_cond)
            & ~traces[COL_FUNC_ID].str.contains(write_cond),
            IOCategory.METADATA.value,
        )
        traces[COL_IO_CAT] = traces[COL_IO_CAT].mask(
            (traces['cat'] == CAT_POSIX)
            & traces[COL_FUNC_ID].str.contains(read_cond)
            & ~traces[COL_FUNC_ID].str.contains(metadata_cond),
            IOCategory.READ.value,
        )
        traces[COL_IO_CAT] = traces[COL_IO_CAT].mask(
            (traces['cat'] == CAT_POSIX)
            & traces[COL_FUNC_ID].str.contains(write_cond)
            & ~traces[COL_FUNC_ID].str.contains(metadata_cond),
            IOCategory.WRITE.value,
        )
        return traces

    def compute_job_time(self, traces: dd.DataFrame) -> float:
        return (traces['te'].max() - traces['ts'].min()) / DFTRACER_TIME_RESOLUTION

    def compute_total_count(self, traces: dd.DataFrame) -> int:
        return (
            traces[(traces['cat'] == CAT_POSIX) & (traces['ts'] > 0)]
            .index.count()
            .persist()
        )



import logging
import socket
from dask.distributed import Client, LocalCluster
from dask_jobqueue import LSFCluster, PBSCluster
from time import sleep
from .config import ClusterConfig, ClusterType, get_working_dir
from .utils.file_utils import ensure_dir


class ClusterManager(object):
    def __init__(self, config: ClusterConfig):
        self.config = config
        self.working_dir = get_working_dir()
        ensure_dir(self.working_dir)
        ensure_dir(f"{self.working_dir}/worker_logs")

    def boot(self):
        self.cluster = self._initialize_cluster()
        self.client = Client(self.cluster)
        if self.config.type != ClusterType.LOCAL:
            self.cluster.scale(self.config.n_workers)
            logging.info(f"Scaling cluster to {self.config.n_workers} nodes")
            self._wait_until_workers_alive()

    def shutdown(self):
        self.client.close()
        self.cluster.close()

    def _initialize_cluster(self):
        dashboard_address = None
        host_name = None
        if self.config.host is not None:
            host_name = f"{self.config.host}"
        elif self.config.type != ClusterType.LOCAL:
            host_name = socket.gethostname()
        if self.config.dashboard_port is not None:
            assert host_name is not None, 'Host address must be specified'
            dashboard_address = f"{host_name}:{self.config.dashboard_port}"
        if self.config.type == ClusterType.LOCAL:
            return LocalCluster(
                host=host_name,
                local_directory=self.config.local_dir,
                memory_limit=self.config.memory,
                n_workers=self.config.n_workers,
                processes=self.config.processes,
                silence_logs=logging.DEBUG if self.config.debug else logging.CRITICAL,
            )
        elif self.config.type == ClusterType.LSF:
            return LSFCluster(
                cores=self.config.n_workers * self.config.n_threads_per_worker,
                death_timeout=self.config.death_timeout,
                job_directives_skip=self.config.job_directives_skip,
                job_extra_directives=self.config.job_extra_directives,
                local_directory=self.config.local_dir,
                memory=f"{self.config.memory}GB",
                processes=self.config.n_workers,
                scheduler_options=dict(
                    dashboard_address=dashboard_address,
                    host=host_name,
                ),
                use_stdin=self.config.use_stdin,
            )
        elif self.config.type == ClusterType.PBS:
            return PBSCluster(
                cores=self.config.n_workers * self.config.n_threads_per_worker,
                death_timeout=self.config.death_timeout,
                job_directives_skip=self.config.job_directives_skip,
                job_extra_directives=self.config.job_extra_directives,
                local_directory=self.config.local_dir,
                log_directory=f"{self.working_dir}/worker_logs",
                memory=f"{self.config.memory}GB",
                processes=self.config.n_workers,
                scheduler_options=dict(
                    dashboard_address=dashboard_address,
                    host=host_name,
                ),
            )

    def _wait_until_workers_alive(self, sleep_seconds=2):
        active_workers = len(self.client.scheduler_info()['workers'])
        while (
            self.client.status == 'running' and active_workers < self.config.n_workers
        ):
            active_workers = len(self.client.scheduler_info()['workers'])
            logging.debug(
                f"Waiting for workers ({active_workers}/{self.config.n_workers})"
            )
            # Try to force cluster to boot workers
            self.cluster._correct_state()
            # Wait
            sleep(sleep_seconds)
        logging.debug('All workers alive')



import abc
import dask

dask.config.set({'dataframe.query-planning-warning': False})
import dask.dataframe as dd
import hashlib
import itertools as it
import json
import logging
import math
import os
from dask.base import compute, unpack_collections
from dask.delayed import Delayed
from dask.distributed import fire_and_forget, get_client, wait
from typing import Callable, Dict, List, Optional, Tuple, Union

from .analysis import THRESHOLD_FUNCTIONS, set_metric_slope
from .analysis_utils import set_file_dir, set_file_pattern, set_proc_name_parts
from .constants import (
    ACC_PAT_SUFFIXES,
    COL_FILE_NAME,
    COL_PROC_NAME,
    DERIVED_MD_OPS,
    EVENT_ATT_REASONS,
    EVENT_COMP_HLM,
    EVENT_COMP_MAIN_VIEW,
    EVENT_COMP_METBD,
    EVENT_COMP_PERS,
    EVENT_DET_BOT,
    EVENT_SAVE_BOT,
    EVENT_SAVE_VIEWS,
    LOGICAL_VIEW_TYPES,
    AccessPattern,
    IOCategory,
)
from .rule_engine import RuleEngine
from .scoring import ViewEvaluator
from .types import (
    AnalysisAccuracy,
    AnalyzerResultType,
    Metric,
    RawStats,
    ViewKey,
    ViewResult,
    ViewType,
)
from .utils.dask_utils import EventLogger, flatten_column_names
from .utils.file_utils import ensure_dir
from .utils.json_encoders import NpEncoder


CHECKPOINT_MAIN_VIEW = '_main_view'
CHECKPOINT_METRIC_BOUNDARIES = '_metric_boundaries'
CHECKPOINT_HLM = '_hlm'
CHECKPOINT_RAW_STATS = '_raw_stats'
CHECKPOINT_VIEW = '_view'
EXTRA_COLS = ['io_cat', 'acc_pat', 'func_id']
HLM_AGG = {
    'time': [sum],
    'count': [sum],
    'size': [min, max, sum],
}
HLM_COLS = {
    'count_sum': 'count',
    'size_sum': 'size',
    'time_sum': 'time',
}
VIEW_AGG = {
    # 'bw': max,
    'count': sum,
    'data_count': sum,
    # 'intensity': max,
    # 'iops': max,
    'size': sum,
    'time': sum,
}
HASH_CHECKPOINT_NAMES = False
WAIT_ENABLED = True


class Analyzer(abc.ABC):
    def __init__(
        self,
        bottleneck_dir: str = "",
        checkpoint: bool = True,
        checkpoint_dir: str = "",
        debug: bool = False,
        time_approximate: bool = True,
        time_granularity: float = 1e6,
        verbose: bool = False,
    ):
        """Initializes the Analyzer instance.

        Args:
            bottleneck_dir: Directory to save identified bottlenecks.
            checkpoint: Whether to enable checkpointing of intermediate results.
            checkpoint_dir: Directory to store checkpoint data.
            debug: Whether to enable debug mode.
            time_approximate: Whether to use approximate time for I/O operations.
            time_granularity: The time granularity for analysis, in microseconds.
            verbose: Whether to enable verbose logging.
        """
        if checkpoint:
            assert checkpoint_dir != '', 'Checkpoint directory must be defined'

        self.bottleneck_dir = bottleneck_dir
        self.checkpoint = checkpoint
        self.checkpoint_dir = checkpoint_dir
        self.debug = debug
        self.time_approximate = time_approximate
        self.time_granularity = time_granularity
        self.verbose = verbose

        # Setup directories
        ensure_dir(self.bottleneck_dir)
        ensure_dir(self.checkpoint_dir)

    def analyze_trace(
        self,
        trace_path: str,
        accuracy: AnalysisAccuracy = 'pessimistic',
        exclude_bottlenecks: List[str] = [],
        exclude_characteristics: List[str] = [],
        logical_view_types: bool = False,
        metrics: List[Metric] = ['iops'],
        percentile: Optional[float] = None,
        threshold: Optional[int] = None,
        view_types: List[ViewType] = ['file_name', 'proc_name', 'time_range'],
    ) -> AnalyzerResultType:
        """Analyzes I/O trace data to identify performance bottlenecks.

        This method orchestrates the entire analysis process, including reading
        trace data, computing various metrics and views, evaluating these views
        to detect bottlenecks, and applying rules to characterize them.

        Args:
            trace_path: Path to the I/O trace file or directory.
            accuracy: The analysis accuracy mode ('optimistic' or 'pessimistic').
            exclude_bottlenecks: A list of bottleneck types to exclude from the analysis.
            exclude_characteristics: A list of I/O characteristics to exclude.
            logical_view_types: Whether to compute views based on logical relationships.
            metrics: A list of metrics to analyze (e.g., 'iops', 'bw', 'time').
            percentile: The percentile to use for identifying critical views.
                        Mutually exclusive with 'threshold'.
            threshold: The threshold value for slope-based bottleneck detection.
                       Mutually exclusive with 'percentile'.
            view_types: A list of view types to compute (e.g., 'file_name', 'proc_name').

        Returns:
            An AnalyzerResultType object containing the analysis results.

        Raises:
            ValueError: If neither 'percentile' nor 'threshold' is defined.
        """
        # Check if both percentile and threshold are none
        if percentile is None and threshold is None:
            raise ValueError('Either percentile or threshold must be defined')
        is_slope_based = threshold is not None

        # Read trace & stats
        traces = self.read_trace(trace_path=trace_path)
        raw_stats = self.read_stats(traces=traces)
        traces = self.postread_trace(traces=traces)

        # Create checkpoint names
        main_view_name = self.get_checkpoint_name(CHECKPOINT_MAIN_VIEW, *sorted(view_types))

        # Check there is a checkpointed main view
        if not self.checkpoint or not self.has_checkpoint(name=main_view_name):
            # Compute high-level metrics
            with EventLogger(key=EVENT_COMP_HLM, message='Compute high-level metrics'):
                hlm = self.restore_view(
                    name=self.get_checkpoint_name(CHECKPOINT_HLM, *sorted(view_types)),
                    fallback=lambda: self.compute_high_level_metrics(
                        traces=traces,
                        view_types=view_types,
                    ),
                )
                hlm = hlm.persist()
                self._wait_all(tasks=hlm)

        # Compute main view
        with EventLogger(key=EVENT_COMP_MAIN_VIEW, message='Compute main view'):
            main_view = self.restore_view(
                name=main_view_name,
                fallback=lambda: self.compute_main_view(
                    hlm=hlm,
                    view_types=view_types,
                ),
            )
            # TODO remove dropped columns
            main_view = main_view.drop(columns=['bw', 'intensity', 'iops', 'att_perf'], errors='ignore').persist()
            self._wait_all(tasks=main_view)

        # return traces, main_view
        return self._analyze_main_view(
            main_view=main_view,
            metrics=metrics,
            percentile=percentile,
            threshold=threshold,
            view_types=view_types,
            is_slope_based=is_slope_based,
            raw_stats=raw_stats,
            exclude_bottlenecks=exclude_bottlenecks,
            exclude_characteristics=exclude_characteristics,
        )

    def read_stats(self, traces: dd.DataFrame) -> RawStats:
        """Computes and restores raw statistics from the trace data.

        Calculates job time and total event count from the traces.
        It attempts to restore these stats from a checkpoint if available,
        otherwise computes them and checkpoints the result.

        Args:
            traces: A Dask DataFrame containing the I/O trace data.

        Returns:
            A RawStats dictionary containing 'job_time', 'time_granularity',
            and 'total_count'.
        """
        job_time = self.compute_job_time(traces=traces)
        total_count = self.compute_total_count(traces=traces)
        raw_stats: RawStats = self.restore_extra_data(
            name=self.get_checkpoint_name(CHECKPOINT_RAW_STATS),
            fallback=lambda: dict(
                job_time=job_time,
                time_granularity=self.time_granularity,
                total_count=total_count,
            ),
        )
        return raw_stats

    def read_trace(self, trace_path: str) -> dd.DataFrame:
        """Reads I/O trace data from the specified path.

        This is an abstract method that must be implemented by subclasses
        to handle specific trace formats.

        Args:
            trace_path: Path to the I/O trace file or directory.

        Returns:
            A Dask DataFrame containing the parsed I/O trace data.

        Raises:
            NotImplementedError: If the subclass does not implement this method.
        """
        raise NotImplementedError

    def postread_trace(self, traces: dd.DataFrame) -> dd.DataFrame:
        """Performs any post-processing on the raw trace data.

        This method can be overridden by subclasses to perform additional
        transformations or filtering on the trace data after it has been read.
        By default, it returns the traces unmodified.

        Args:
            traces: A Dask DataFrame containing the I/O trace data.

        Returns:
            A Dask DataFrame with any post-processing applied.
        """
        return traces

    def compute_job_time(self, traces: dd.DataFrame) -> float:
        """Computes the total job execution time from the traces.

        Args:
            traces: A Dask DataFrame containing the I/O trace data,
                    expected to have 'tstart' and 'tend' columns.

        Returns:
            The total job time as a float.
        """
        return traces['tend'].max() - traces['tstart'].min()

    def compute_total_count(self, traces: dd.DataFrame) -> int:
        """Computes the total number of I/O events in the traces.

        Args:
            traces: A Dask DataFrame containing the I/O trace data.

        Returns:
            The total count of I/O events as an integer.
        """
        return traces.index.count().persist()

    def compute_high_level_metrics(
        self,
        traces: dd.DataFrame,
        view_types: list,
        partition_size: str = '256MB',
    ) -> dd.DataFrame:
        """Computes high-level metrics by aggregating trace data.

        Groups the trace data by the specified view types and extra columns
        (io_cat, acc_pat, func_id) and aggregates metrics like time, count, and size.

        Args:
            traces: A Dask DataFrame containing the I/O trace data.
            view_types: A list of column names to group by for aggregation.
            partition_size: The desired partition size for the resulting Dask DataFrame.

        Returns:
            A Dask DataFrame containing the computed high-level metrics.
        """
        # Add `io_cat`, `acc_pat`, and `func_id` to groupby
        groupby = list(view_types)
        groupby.extend(EXTRA_COLS)

        # Compute high-level metrics
        # hlm = traces \
        #     .groupby(groupby) \
        #     .agg(HLM_AGG, split_out=traces.npartitions) \
        #     .persist() \
        #     .reset_index() \
        #     .repartition(partition_size) \
        #     .persist()
        # hlm = traces \
        #     .groupby(groupby) \
        #     .agg(HLM_AGG, split_out=8) \
        #     .reset_index() \
        #     .persist()
        hlm = (
            traces.groupby(groupby)
            .agg(HLM_AGG, split_out=math.ceil(math.sqrt(traces.npartitions)))
            .persist()
            .reset_index()
            .repartition(partition_size=partition_size)
        )
        hlm = flatten_column_names(hlm)
        return hlm.rename(columns=HLM_COLS).persist()

    def compute_main_view(
        self,
        hlm: dd.DataFrame,
        view_types: List[ViewType],
        partition_size: str = '256MB',
    ) -> dd.DataFrame:
        """Computes the main aggregated view from high-level metrics.

        This method takes the high-level metrics, sets derived columns,
        and then groups by the specified view_types to create a primary
        aggregated view of the I/O performance data.

        Args:
            hlm: A Dask DataFrame containing high-level metrics.
            view_types: A list of view types to group by for the main view.
            partition_size: The desired partition size for the resulting Dask DataFrame.

        Returns:
            A Dask DataFrame representing the main aggregated view.
        """
        # Set derived columns
        hlm = self._set_derived_columns(ddf=hlm)
        hlm_agg = {col: sum for col in hlm.columns if col not in EXTRA_COLS}
        for view_type in view_types:
            if view_type in hlm_agg:
                hlm_agg.pop(view_type)
        # Compute agg_view
        main_view = (
            hlm.groupby(list(view_types))
            .agg(hlm_agg, split_out=hlm.npartitions)
            .persist()
            .repartition(partition_size=partition_size)
        )
        # main_view = hlm \
        #     .drop(columns=EXTRA_COLS) \
        #     .groupby(view_types) \
        #     .sum() \
        #     .persist() \
        #     .repartition(partition_size)
        # Set hashed ids
        # main_view['id'] = main_view.index.map(set_id)
        # main_view['id'] = main_view.index.map(hash)
        # Return main_view
        return main_view

    def compute_metric_boundaries(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        view_types: List[ViewType],
    ) -> Dict[Metric, dd.core.Scalar]:
        """Computes the upper boundary for each specified metric.

        For metrics like 'iops' or 'time', it calculates the maximum time
        either per process (if 'proc_name' is in view_types) or the total sum.
        Other metrics like 'bw' are currently passed through.

        Args:
            main_view: The main aggregated Dask DataFrame.
            metrics: A list of metrics for which to compute boundaries.
            view_types: A list of view types present in the main_view.

        Returns:
            A dictionary mapping each metric to its computed boundary (a Dask Scalar).
        """
        metric_boundaries = {}
        for metric in metrics:
            metric_boundary = None
            if metric == 'iops' or metric == 'time':
                if COL_PROC_NAME in view_types:
                    metric_boundary = main_view.groupby([COL_PROC_NAME]).sum()['time'].max().persist()
                else:
                    metric_boundary = main_view['time'].sum().persist()
            elif metric == 'bw':
                pass
            metric_boundaries[metric] = metric_boundary
        return metric_boundaries

    def compute_views(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        metric_boundaries: Dict[Metric, dd.core.Scalar],
        percentile: Optional[float],
        threshold: Optional[int],
        view_types: List[ViewType],
    ):
        """Computes multifaceted views for each specified metric.

        Iterates through all permutations of view_types for each metric,
        generating different "perspectives" on the data. Each perspective
        is a ViewResult, containing the filtered data and critical items.

        Args:
            main_view: The main aggregated Dask DataFrame.
            metrics: A list of metrics to compute views for.
            metric_boundaries: A dictionary of precomputed metric boundaries.
            percentile: The percentile used to identify critical items in views.
            threshold: The threshold value for slope-based critical item identification.
            view_types: A list of base view types to permute for creating views.

        Returns:
            A dictionary where keys are metrics and values are dictionaries
            mapping ViewKey to ViewResult.
        """
        # Keep view results
        view_results = {}

        # Compute multifaceted views for each metric
        for metric in metrics:
            view_results[metric] = {}

            for view_key in self.view_permutations(view_types=view_types):
                view_type = view_key[-1]
                parent_view_key = view_key[:-1]

                parent_view_result = view_results[metric].get(parent_view_key, None)
                parent_records = main_view if parent_view_result is None else parent_view_result.records

                view_result = self.compute_view(
                    metrics=metrics,
                    metric=metric,
                    metric_boundary=metric_boundaries[metric],
                    percentile=percentile,
                    records=parent_records,
                    threshold=threshold,
                    view_key=view_key,
                    view_type=view_type,
                )

                view_results[metric][view_key] = view_result

        return view_results

    def compute_logical_views(
        self,
        main_view: dd.DataFrame,
        metric_boundaries: Dict[Metric, dd.core.Scalar],
        metrics: List[Metric],
        percentile: Optional[float],
        threshold: Optional[int],
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
        view_types: List[ViewType],
    ):
        """Computes views based on predefined logical relationships in the data.

        This method extends the existing view_results by adding new views
        derived from logical columns (e.g., file directory from file name).

        Args:
            main_view: The main aggregated Dask DataFrame.
            metric_boundaries: A dictionary of precomputed metric boundaries.
            metrics: A list of metrics to compute logical views for.
            percentile: The percentile used to identify critical items in views.
            threshold: The threshold value for slope-based critical item identification.
            view_results: The existing dictionary of computed views to be updated.
            view_types: A list of base view types available in the main_view.

        Returns:
            The updated view_results dictionary including the computed logical views.
        """
        for metric in metrics:
            for view_key in LOGICAL_VIEW_TYPES:
                view_type = view_key[-1]
                parent_view_key = view_key[:-1]
                parent_view_type = parent_view_key[0]

                if parent_view_type not in view_types:
                    continue

                parent_view_result = view_results[metric].get(parent_view_key, None)
                parent_records = main_view if parent_view_result is None else parent_view_result.records

                if view_type not in parent_records.columns:
                    parent_records = self._set_logical_columns(
                        view=parent_records,
                        view_types=[parent_view_type],
                    )

                view_result = self.compute_view(
                    metrics=metrics,
                    metric=metric,
                    metric_boundary=metric_boundaries[metric],
                    records=parent_records,
                    percentile=percentile,
                    threshold=threshold,
                    view_key=view_key,
                    view_type=view_type,
                )

                view_results[metric][view_key] = view_result

        return view_results

    def compute_view(
        self,
        metrics: List[Metric],
        metric: Metric,
        metric_boundary: dd.core.Scalar,
        percentile: Optional[float],
        records: dd.DataFrame,
        threshold: Optional[int],
        view_key: ViewKey,
        view_type: str,
    ) -> ViewResult:
        """Computes a single view based on the provided parameters.

        This involves restoring a view from a checkpoint or computing it,
        then filtering it to identify critical items based on percentile or threshold.

        Args:
            metrics: The list of all metrics being analyzed.
            metric: The specific metric for this view.
            metric_boundary: The precomputed boundary for the current metric.
            percentile: The percentile to identify critical items.
            records: The Dask DataFrame (parent records) to compute the view from.
            threshold: The threshold for slope-based critical item identification.
            view_key: The key identifying this specific view.
            view_type: The primary dimension/column for this view.

        Returns:
            A ViewResult object containing the computed view, critical items,
            and filtered records.
        """
        # Restore view
        view = self.restore_view(
            name=self.get_checkpoint_name(CHECKPOINT_VIEW, metric, *list(view_key)),
            fallback=lambda: self._compute_view(
                records=records,
                view_type=view_type,
                metrics=metrics,
                metric=metric,
                metric_boundary=metric_boundary,
            ),
            write_to_disk=False,
        )

        # Filter by slope
        critical_view = self._compute_critical_view(
            view=view,
            metric=metric,
            percentile=percentile,
            threshold=threshold,
        )

        indices = critical_view.index.unique()

        # Find filtered records
        records = records.query(f"{view_type} in @indices", local_dict={'indices': indices}).persist()

        # Return views & normalization data
        return ViewResult(
            critical_view=critical_view,
            metric=metric,
            records=records,
            view=view,
            view_type=view_type,
        )

    def get_checkpoint_name(self, *args) -> str:
        """Generates a standardized name for a checkpoint.

        Joins the provided arguments with underscores. If HASH_CHECKPOINT_NAMES
        is True, it returns an MD5 hash of the name.

        Args:
            *args: String components to form the checkpoint name.

        Returns:
            A string representing the checkpoint name.
        """
        checkpoint_name = "_".join(args)
        if HASH_CHECKPOINT_NAMES:
            return hashlib.md5(checkpoint_name.encode("utf-8")).hexdigest()
        return checkpoint_name

    def get_checkpoint_path(self, name: str) -> str:
        """Constructs the full path for a given checkpoint name.

        Args:
            name: The name of the checkpoint.

        Returns:
            The absolute path to the checkpoint directory/file.
        """
        return f"{self.checkpoint_dir}/{name}"

    def has_checkpoint(self, name: str) -> bool:
        """Checks if a checkpoint with the given name exists.

        A checkpoint is considered to exist if its `_metadata` file is present.

        Args:
            name: The name of the checkpoint.

        Returns:
            True if the checkpoint exists, False otherwise.
        """
        checkpoint_path = self.get_checkpoint_path(name=name)
        return os.path.exists(f"{checkpoint_path}/_metadata")

    def restore_extra_data(self, name: str, fallback: Callable[[], dict], force=False, persist=False) -> dict:
        """Restores extra (non-DataFrame) data from a JSON checkpoint.

        If checkpointing is enabled and the checkpoint file exists (unless 'force'
        is True), it loads the data from the JSON file. Otherwise, it calls the
        'fallback' function to compute the data and then stores it asynchronously.

        Args:
            name: The name of the checkpoint.
            fallback: A callable function that returns the data if not found or forced.
            force: If True, forces recomputation even if a checkpoint exists.
            persist: (Currently unused in the method body, but part of signature)

        Returns:
            A dictionary containing the restored or computed data.
        """
        if self.checkpoint:
            data_path = f"{self.get_checkpoint_path(name=name)}.json"
            if force or not os.path.exists(data_path):
                data = fallback()
                fire_and_forget(
                    get_client().submit(
                        self.store_extra_data,
                        data=get_client().submit(compute, data),
                        data_path=data_path,
                    )
                )
                return data
            with open(data_path, 'r') as f:
                return json.load(f)
        return fallback()

    def restore_view(
        self,
        name: str,
        fallback: Callable[[], dd.DataFrame],
        force=False,
        write_to_disk=True,
    ) -> dd.DataFrame:
        """Restores a Dask DataFrame view from a Parquet checkpoint.

        If checkpointing is enabled and the checkpoint exists (unless 'force' is True),
        it reads the DataFrame from the Parquet store. Otherwise, it calls the
        'fallback' function to compute the DataFrame. If 'write_to_disk' is True,
        the computed DataFrame is then stored as a checkpoint.

        Args:
            name: The name of the checkpoint.
            fallback: A callable function that returns the DataFrame if not found or forced.
            force: If True, forces recomputation even if a checkpoint exists.
            write_to_disk: If True, saves the computed view to disk if it was recomputed.

        Returns:
            A Dask DataFrame representing the restored or computed view.
        """
        if self.checkpoint:
            view_path = self.get_checkpoint_path(name=name)
            if force or not self.has_checkpoint(name=name):
                view = fallback()
                if not write_to_disk:
                    return view
                self.store_view(name=name, view=view)
                get_client().cancel(view)
            return dd.read_parquet(view_path)
        return fallback()

    def save_bottlenecks(self, bottlenecks: dd.DataFrame, partition_size='64MB'):
        """Saves the identified bottlenecks to Parquet files.

        The bottlenecks DataFrame is repartitioned and then written to the
        `bottleneck_dir` specified during Analyzer initialization.

        Args:
            bottlenecks: A Dask DataFrame containing the identified bottlenecks.
            partition_size: The desired partition size for the output Parquet files.

        Returns:
            The result of the Dask `to_parquet` operation (typically None or a Dask future).
        """
        return bottlenecks.repartition(partition_size=partition_size).to_parquet(
            self.bottleneck_dir, compute=True, write_metadata_file=True
        )

    @staticmethod
    def store_extra_data(data: Tuple[Dict], data_path: str):
        """Saves extra (non-DataFrame) data to a JSON file.

        This static method is typically used by Dask workers to persist data.

        Args:
            data: A tuple containing a single dictionary of data to be saved.
            data_path: The full path to the JSON file where data will be stored.
        """
        with open(data_path, 'w') as f:
            return json.dump(data[0], f, cls=NpEncoder)

    def store_view(self, name: str, view: dd.DataFrame, compute=True, partition_size='64MB'):
        """Stores a Dask DataFrame view to a Parquet checkpoint.

        The view DataFrame is repartitioned and then written to a subdirectory
        named `name` within the `checkpoint_dir`.

        Args:
            name: The name of the checkpoint.
            view: The Dask DataFrame to store.
            compute: Whether to compute the DataFrame before writing (Dask default is True).
            partition_size: The desired partition size for the output Parquet files.

        Returns:
            The result of the Dask `to_parquet` operation.
        """
        return view.repartition(partition_size=partition_size).to_parquet(
            self.get_checkpoint_path(name=name),
            compute=compute,
            write_metadata_file=True,
        )

    @staticmethod
    def view_permutations(view_types: List[ViewType]):
        """Generates all permutations of view_types for creating multifaceted views.

        For a list of view_types [vt1, vt2, vt3], it will generate permutations
        of length 1, 2, and 3, e.g., (vt1,), (vt2,), (vt1, vt2), (vt2, vt1), ...

        Args:
            view_types: A list of ViewType elements.

        Returns:
            An iterator yielding tuples, where each tuple is a permutation of view_types.
        """

        def _iter_permutations(r: int):
            return it.permutations(view_types, r + 1)

        return it.chain.from_iterable(map(_iter_permutations, range(len(view_types))))

    def _analyze_main_view(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        percentile: Optional[float],
        threshold: Optional[int],
        view_types: List[ViewType],
        is_slope_based: bool,
        raw_stats: RawStats,
        exclude_bottlenecks: List[str],
        exclude_characteristics: List[str],
        logical_view_types: bool = False,
    ):
        # Compute upper bounds
        with EventLogger(key=EVENT_COMP_METBD, message='Compute metric boundaries'):
            metric_boundaries = self.restore_extra_data(
                name=self.get_checkpoint_name(
                    CHECKPOINT_METRIC_BOUNDARIES,
                    *sorted(metrics),
                    *sorted(view_types),
                ),
                fallback=lambda: self.compute_metric_boundaries(
                    main_view=main_view,
                    metrics=metrics,
                    view_types=view_types,
                ),
            )
            self._wait_all(tasks=metric_boundaries)

        # Compute views
        with EventLogger(key=EVENT_COMP_PERS, message='Compute perspectives'):
            view_results = self.compute_views(
                main_view=main_view,
                metric_boundaries=metric_boundaries,
                metrics=metrics,
                percentile=percentile,
                threshold=threshold,
                view_types=view_types,
            )
            if logical_view_types:
                logical_view_results = self.compute_logical_views(
                    main_view=main_view,
                    metric_boundaries=metric_boundaries,
                    metrics=metrics,
                    percentile=percentile,
                    threshold=threshold,
                    view_results=view_results,
                    view_types=view_types,
                )
                view_results.update(logical_view_results)
            self._wait_all(tasks=view_results)

        evaluated_views, bottlenecks, bottleneck_rules, characteristics = self._handle_view_results(
            main_view=main_view,
            metrics=metrics,
            metric_boundaries=metric_boundaries,
            view_results=view_results,
            is_slope_based=is_slope_based,
            raw_stats=raw_stats,
            exclude_bottlenecks=exclude_bottlenecks,
            exclude_characteristics=exclude_characteristics,
        )

        # Return result
        return AnalyzerResultType(
            bottleneck_dir=self.bottleneck_dir,
            bottleneck_rules=bottleneck_rules,
            characteristics=characteristics,
            evaluated_views=evaluated_views,
            main_view=main_view,
            metric_boundaries=metric_boundaries,
            raw_stats=raw_stats,
            view_results=view_results,
        )

    def _compute_critical_view(
        self,
        view: dd.DataFrame,
        metric: Metric,
        percentile: Optional[float],
        threshold: Optional[int],
    ):
        """Computes the critical view based on the specified metric.

        This method filters the view DataFrame to identify critical items
        based on either a percentile or a threshold value.

        Args:
            view: The Dask DataFrame representing the view to be filtered.
            metric: The specific metric used for filtering.
            percentile: The percentile to identify critical items.
            threshold: The threshold for slope-based critical item identification.

        Returns:
            A Dask DataFrame containing the filtered critical view.
        """
        if percentile is not None and percentile > 0:
            return view.query(
                f"{metric}_pth >= @percentile",
                local_dict={'percentile': percentile},
            ).persist()
        elif threshold is not None and threshold > 0:
            corrected_threshold = THRESHOLD_FUNCTIONS[metric](threshold)
            return view.query(
                f"{metric}_slope <= @threshold",
                local_dict={'threshold': corrected_threshold},
            ).persist()
        return view

    def _compute_view(
        self,
        records: dd.DataFrame,
        view_type: str,
        metrics: List[Metric],
        metric: Metric,
        metric_boundary: dd.core.Scalar,
    ) -> dd.DataFrame:
        view_types = records.index._meta.names

        non_proc_agg_dict = self._get_agg_dict(
            for_view_type=view_type,
            view_columns=records.columns,
            view_types=view_types,
            is_proc=False,
        )
        proc_agg_dict = self._get_agg_dict(
            for_view_type=view_type,
            view_columns=records.columns,
            view_types=view_types,
            is_proc=True,
        )

        # Check view type
        if view_type is not COL_PROC_NAME and COL_PROC_NAME in view_types:
            view = (
                records.reset_index()
                .groupby([view_type, COL_PROC_NAME])
                .agg(non_proc_agg_dict)
                .groupby([view_type])
                .agg(proc_agg_dict)
            )
        else:
            view = records.reset_index().groupby([view_type]).agg(non_proc_agg_dict)

        # Set metric slope
        view = view.map_partitions(
            set_metric_slope,
            metrics=metrics,
            metric=metric,
            metric_boundary=metric_boundary,
        )

        # Return view
        return view

    @staticmethod
    def _get_agg_dict(
        for_view_type: ViewType,
        view_columns: List[str],
        view_types: List[ViewType],
        is_proc=False,
    ):
        if is_proc:
            agg_dict = {col: max if 'time' in col else sum for col in view_columns}
        else:
            agg_dict = {col: sum for col in view_columns}

        # agg_dict['bw'] = max
        # agg_dict['intensity'] = max
        # agg_dict['iops'] = max
        agg_dict['size_min'] = min
        agg_dict['size_max'] = max

        unwanted_agg_cols = ['id', for_view_type]
        for agg_col in unwanted_agg_cols:
            if agg_col in agg_dict:
                agg_dict.pop(agg_col)

        return agg_dict

    def _handle_view_results(
        self,
        main_view: dd.DataFrame,
        metrics: List[Metric],
        metric_boundaries: Dict[Metric, dd.core.Scalar],
        view_results: Dict[Metric, Dict[ViewKey, ViewResult]],
        is_slope_based: bool,
        raw_stats: RawStats,
        exclude_bottlenecks: List[str],
        exclude_characteristics: List[str],
    ):
        if self.checkpoint:
            with EventLogger(key=EVENT_SAVE_VIEWS, message='Checkpoint views', level=logging.DEBUG):
                view_checkpoint_tasks = []
                for metric, views in view_results.items():
                    for view_key, view_result in views.items():
                        view_checkpoint_name = self.get_checkpoint_name(CHECKPOINT_VIEW, metric, *list(view_key))
                        if not self.has_checkpoint(name=view_checkpoint_name):
                            view_checkpoint_tasks.append(
                                self.store_view(
                                    name=view_checkpoint_name,
                                    view=view_result.view,
                                    compute=True,
                                )
                            )
                self._wait_all(tasks=view_checkpoint_tasks)

        # Evaluate views
        view_evaluator = ViewEvaluator()
        with EventLogger(key=EVENT_DET_BOT, message='Detect I/O bottlenecks'):
            evaluated_views = view_evaluator.evaluate_views(
                metric_boundaries=metric_boundaries,
                metrics=metrics,
                view_results=view_results,
                is_slope_based=is_slope_based,
            )
            self._wait_all(tasks=evaluated_views)

        # Execute rules
        rule_engine = RuleEngine(rules={}, raw_stats=raw_stats, verbose=self.verbose)
        with EventLogger(key=EVENT_ATT_REASONS, message='Attach reasons to I/O bottlenecks'):
            characteristics = rule_engine.process_characteristics(
                exclude_characteristics=exclude_characteristics,
                main_view=main_view,
                view_results=view_results,
            )
            bottlenecks, bottleneck_rules = rule_engine.process_bottlenecks(
                evaluated_views=evaluated_views,
                exclude_bottlenecks=exclude_bottlenecks,
                group_behavior=False,
                metric_boundaries=metric_boundaries,
            )
            self._wait_all(tasks=bottlenecks)

        with EventLogger(key=EVENT_SAVE_BOT, message='Save I/O bottlenecks', level=logging.DEBUG):
            self.save_bottlenecks(bottlenecks=bottlenecks)

        return evaluated_views, bottlenecks, bottleneck_rules, characteristics

    def _set_derived_columns(self, ddf: dd.DataFrame):
        # Derive `io_cat` columns
        for col in ['time', 'size', 'count']:
            for io_cat in list(IOCategory):
                col_name = f"{io_cat.name.lower()}_{col}"
                ddf[col_name] = 0.0 if col == 'time' else 0
                ddf[col_name] = ddf[col_name].mask(ddf['io_cat'] == io_cat.value, ddf[col])
        for io_cat in list(IOCategory):
            min_name, max_name = (
                f"{io_cat.name.lower()}_min",
                f"{io_cat.name.lower()}_max",
            )
            ddf[min_name] = 0
            ddf[max_name] = 0
            ddf[min_name] = ddf[min_name].mask(ddf['io_cat'] == io_cat.value, ddf['size_min'])
            ddf[max_name] = ddf[max_name].mask(ddf['io_cat'] == io_cat.value, ddf['size_max'])
        # Derive `data` columns
        ddf['data_count'] = ddf['write_count'] + ddf['read_count']
        ddf['data_size'] = ddf['write_size'] + ddf['read_size']
        ddf['data_time'] = ddf['write_time'] + ddf['read_time']
        # Derive `acc_pat` columns
        for col_suffix, col_value in zip(ACC_PAT_SUFFIXES, ['data_time', 'data_size', 'data_count']):
            for acc_pat in list(AccessPattern):
                col_name = f"{acc_pat.name.lower()}_{col_suffix}"
                ddf[col_name] = 0.0 if col_suffix == 'time' else 0
                ddf[col_name] = ddf[col_name].mask(ddf['acc_pat'] == acc_pat.value, ddf[col_value])
        # Derive metadata operation columns
        for col in ['time', 'count']:
            for md_op in DERIVED_MD_OPS:
                col_name = f"{md_op}_{col}"
                ddf[col_name] = 0.0 if col == 'time' else 0
                if md_op in ['close', 'open']:
                    ddf[col_name] = ddf[col_name].mask(
                        ddf['func_id'].str.contains(md_op) & ~ddf['func_id'].str.contains('dir'),
                        ddf[col],
                    )
                else:
                    ddf[col_name] = ddf[col_name].mask(ddf['func_id'].str.contains(md_op), ddf[col])
        # Cast columns to correct types
        for col in ddf.columns:
            if col.endswith('_size') or col.endswith('_count'):
                ddf[col] = ddf[col].astype('int64')
            elif col.endswith('_time'):
                ddf[col] = ddf[col].astype('float64')
        # Return ddf
        return ddf

    def _set_logical_columns(self, view: dd.DataFrame, view_types: List[ViewType]) -> dd.DataFrame:
        # Check if view types include `proc_name`
        if COL_PROC_NAME in view_types:
            view = view.map_partitions(set_proc_name_parts)

        # Check if view types include `file_name`
        if COL_FILE_NAME in view_types:
            view = view.map_partitions(set_file_dir).map_partitions(set_file_pattern)

        return view

    @staticmethod
    def _wait_all(tasks: Union[dd.DataFrame, Delayed, dict]):
        if WAIT_ENABLED:
            if isinstance(tasks, dd.DataFrame):
                _ = wait(tasks)
            else:
                all_tasks, _ = unpack_collections(tasks)
                _ = wait(all_tasks)
